"""
PC-BASIC - devices.disk
Disk Devices

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

##############################################################################
# Disk devices

import os
import re
import io
import sys
import errno
import ntpath
import logging

from ...compat import text_type, add_str
from ...compat import get_short_pathname, get_free_bytes, is_hidden, iterchar, random_id
from ...compat import is_readable_text_stream, is_writable_text_stream

from ..base import error
from ..base.tokens import ALPHANUMERIC
from ..codepage import CONTROL
from .. import values
from . import devicebase
from .diskfiles import BinaryFile, TextFile, RandomFile, Locks


# GW-BASIC FILE CONTROL BLOCK structure:
# source: IBM Basic reference 1982 (for BASIC-C, BASIC-D, BASIC-A) appendix I-5
# byte  length  description
# 0     1       Mode 1-input 2-output 4-random 16-append
# 1     38      MSDOS FCB:
# ------------------------
# 0     1       Drive number (0=Default, 1=A, etc)
# 01h   8       blank-padded file name
# 09h   3       blank-padded file extension
# 0Ch   2       current block number
# 0Eh   2       logical record size
# 10h   4       file size
# 14h   2       date of last write (see #01666 at AX=5700h)
# 16h   2       time of last write (see #01665 at AX=5700h) (DOS 1.1+)
# 18h   8       reserved (see #01347,#01348,#01349,#01350,#01351)
# 20h   1       record within current block
# 21h   4       random access record number (if record size is > 64 bytes, high byte is omitted)
# ------------------------
# 39    2       For seq files, the number of sectors read or written. For random files, 1+the last record read or written
# 41    1       Number of bytes in sector when read or written
# 42    1       Number of bytes left in input buffer
# 43    3       (reserved)
# 46    1       Drive number 0=A, 1=B, 248=LPT3,2 250=COM2,1 252=CAS1 253=LPT1 254=SCRN 255=KYBD
# 47    1       Device width
# 48    1       Position in buffer for PRINT#
# 49    1       internal use during LOAD and SAVE
# 50    1       Output position used during tab expansion
# 51    128     Physical data buffer for transfer between DOS and BASIC. Can examine this data in seq I/O mode.
# 179   2       Variable length record size, default 128
# 181   2       Current physical record number
# 183   2       Current logical record number
# 185   1       (reserved)
# 186   2       Position for PRINT#, INPUT#, WRITE#
# 188   n       FIELD buffer, default length 128 (given by /s:n)


# translate os error codes to BASIC error codes
OS_ERROR = {
    # file not found
    errno.ENOENT: error.FILE_NOT_FOUND,
    errno.EISDIR: error.FILE_NOT_FOUND,
    errno.ENOTDIR: error.FILE_NOT_FOUND,
    # permission denied
    errno.EAGAIN: error.PERMISSION_DENIED,
    errno.EACCES: error.PERMISSION_DENIED,
    errno.EBUSY: error.PERMISSION_DENIED,
    errno.EROFS: error.PERMISSION_DENIED,
    errno.EPERM: error.PERMISSION_DENIED,
    # disk full
    errno.ENOSPC: error.DISK_FULL,
    # disk not ready
    errno.ENXIO: error.DISK_NOT_READY,
    errno.ENODEV: error.DISK_NOT_READY,
    # device io error
    errno.EIO: error.DEVICE_IO_ERROR,
    # path/file access error
    errno.EEXIST: error.PATH_FILE_ACCESS_ERROR,
    errno.ENOTEMPTY: error.PATH_FILE_ACCESS_ERROR,
}

# allowable characters in DOS file name
# GW-BASIC also allows 0x7F and up, but replaces accented chars with unaccented
# based on CHCP code page, which may differ from display codepage in COUNTRY.SYS
# this is complex and leads to unpredictable results depending on host platform.
ALLOWABLE_CHARS = set(ALPHANUMERIC + b" !#$%&'()-@^_`{}~")

# posix access modes for BASIC modes INPUT, OUTPUT, RANDOM, APPEND
ACCESS_MODES = {b'I': 'r', b'O': 'w', b'R': 'r+', b'A': 'a'}

# aliases for the utf-8 encoding
UTF_8 = ('utf_8', 'utf-8', 'utf', 'u8', 'utf8')


##############################################################################
# exception handling

def safe(fnname, *fnargs):
    """Execute OS function and handle errors."""
    try:
        return fnname(*fnargs)
    except EnvironmentError as e:
        handle_oserror(e)

def handle_oserror(e):
    """Translate OS and I/O exceptions to BASIC errors."""
    try:
        basic_err = OS_ERROR[e.errno]
    except KeyError:
        logging.error(u'Unmapped environment exception: %s', e.errno)
        basic_err = error.DEVICE_IO_ERROR
    raise error.BASICError(basic_err)


##############################################################################
# dos path and filename utilities

def dos_splitext(dos_name):
    """Return trunk and extension excluding the dot."""
    # take whatever comes after first dot as extension
    # and whatever comes before first dot as trunk
    # differs from ntpath.splitext:
    # - does not include . in extension; no extension equals ending in .
    # - dotfiles are trunks starting with . in ntpath but extensions here.
    assert isinstance(dos_name, bytes), type(dos_name)
    elements = dos_name.split(b'.', 1)
    if len(elements) == 1:
        trunk, ext = elements[0], b''
    else:
        trunk, ext = elements
    return trunk, ext

def dos_normalise_name(dos_name):
    """Convert dosname into bytes uppercase 8.3."""
    # a normalised DOS-name is all-uppercase, no leading or trailing spaces, and
    # 1) . or ..; or
    # 2) 0--8 allowable characters followed by one dot followed by 0--3 allowable characters; or
    # 3) 1--8 allowable characters with no dots
    #
    # don't try to split special directory names
    if dos_name in (b'.', b'..'):
        return dos_name
    # convert to all uppercase
    dos_name = dos_name.upper()
    # split into trunk and extension
    trunk, ext = dos_splitext(dos_name)
    # truncate to 8.3
    trunk, ext = trunk[:8], ext[:3]
    if ext:
        ext = b'.' + ext
    norm_name = trunk + ext
    return norm_name

def dos_is_legal_name(dos_name):
    """Check if a (bytes) name is a legal DOS name."""
    assert isinstance(dos_name, bytes), type(dos_name)
    if dos_name in (b'.', b'..'):
        return True
    trunk, ext = dos_splitext(dos_name)
    return (
            # enforce lengths
            (len(trunk) <= 8 and len(ext) <= 3) and
            # no leading or trailing spaces
            (trunk == trunk.strip() and ext == ext.strip()) and
            # enforce allowable characters
            ((set(trunk) | set(ext)) <= ALLOWABLE_CHARS)
        )

def dos_to_native_name(native_path, dosname, isdir):
    """Find a matching native file name for a given normalised DOS name."""
    try:
        uni_name = dosname.decode('ascii')
    except UnicodeDecodeError:
        # non-ascii characters are not allowable for DOS filenames, no match
        return None
    # check if the 8.3 uppercase exists, prefer if so
    if istype(native_path, uni_name, isdir):
        return uni_name
    # otherwise try in lexicographic order
    try:
        all_names = os.listdir(native_path)
    except EnvironmentError:
        # report no match if listdir fails
        return None
    for f in sorted(all_names):
        # we won't match non-ascii anyway
        try:
            ascii_name = f.encode('ascii')
        except UnicodeEncodeError:
            continue
        # don't match long names or non-legal dos names
        if dos_is_legal_name(ascii_name):
            try_name = dos_normalise_name(ascii_name)
            if try_name == dosname and istype(native_path, f, isdir):
                return f
    return None

def dos_name_matches(name, mask):
    """Whether native name element matches DOS wildcard mask."""
    # convert wildcard mask to regexp
    regexp = b'\\A'
    for c in iterchar(mask.upper()):
        if c == b'?':
            regexp += b'.'
        elif c == b'*':
            # we won't need to match newlines, so dot is fine
            regexp += b'.*'
        else:
            regexp += re.escape(c)
    regexp += b'\\Z'
    cregexp = re.compile(regexp)
    return cregexp.match(name.upper()) is not None


##############################################################################
# disk device mapped to native filesystem

class DiskDevice(object):
    """Disk device (A:, B:, C:, ...) """

    allowed_modes = b'IOR'

    def __init__(self, letter, path, cwd, codepage, text_mode, soft_linefeed):
        """Initialise a disk device."""
        assert isinstance(cwd, text_type), type(cwd)
        # DOS drive letter
        self.letter = letter
        # mount root: this is a native filesystem path, using os.sep
        self._native_root = path
        # code page for file system names and text file conversion
        self._codepage = codepage
        # current native working directory on this drive
        self._native_cwd = u''
        if self._native_root:
            self._native_cwd = cwd
        # locks are drive-specific
        self._locks = Locks()
        # text file settings
        # use a BOM on input and output, but not append
        if not text_mode:
            text_mode = ''
        if text_mode.lower() in UTF_8:
            text_mode = 'utf-8-sig'
        self._text_mode = text_mode
        self._soft_linefeed = soft_linefeed

    def close(self):
        """Close disk device."""
        pass

    def available(self):
        """Device is available."""
        return True

    @staticmethod
    def _is_text_file(filetype, mode):
        """Determine if a filetype and mode refer to a text file."""
        return filetype in (b'A', b'D') and mode in (b'O', b'A', b'I')

    def _create_file_object(
            self, fhandle, filetype, mode, number=0,
            field=None, reclen=128, seg=0, offset=0, length=0
        ):
        """Create disk file object of requested type."""
        # determine file type if needed
        if len(filetype) > 1:
            assert mode == b'I', 'file type can only be detected on input files'
            # read magic
            first = fhandle.read(1)
            fhandle.seek(0)
            try:
                filetype_found = devicebase.MAGIC_TO_TYPE[first]
                if filetype_found not in filetype:
                    raise error.BASICError(error.BAD_FILE_MODE)
                filetype = filetype_found
            except KeyError:
                filetype = b'A'
        # for text & ascii-program files, not for random-access files
        if self._is_text_file(filetype, mode):
            # access non-raw text files as text stream
            if self._text_mode and (
                    (mode == b'I' and not is_readable_text_stream(fhandle))
                    or (mode in (b'O', b'A') and not is_writable_text_stream(fhandle))
                ):
                # preserve original newlines on reading and writing
                fhandle = io.TextIOWrapper(
                    fhandle, encoding=self._text_mode, errors='replace', newline=''
                )
            # wrap unicode or bytes native stream so that we always read/write codepage bytes
            if mode in (b'O', b'A'):
                # if the input stream is unicode: decode codepage bytes
                fhandle = self._codepage.wrap_output_stream(fhandle, preserve=CONTROL+(b'\x1A',))
            else: #if mode == b'I':
                # if the input stream is unicode: encode codepage bytes
                # replace newlines with \r in text mode
                fhandle = self._codepage.wrap_input_stream(
                    fhandle, replace_newlines=not self._soft_linefeed
                )
            # ascii program file; or data file for input, output, append
            return TextFile(fhandle, filetype, number, mode, self._locks)
        elif filetype in (b'B', b'P', b'M'):
            # binary [B]LOAD, [B]SAVE
            return BinaryFile(fhandle, filetype, number, mode, seg, offset, length, self._locks)
        elif filetype == b'D' and mode == b'R':
            # data file for random
            return RandomFile(fhandle, number, field, reclen, self._locks)
        else:
            # incorrect file type requested
            msg = b'Incorrect file type %s requested for mode %s' % (filetype, mode)
            raise ValueError(msg)

    def open(
            self, number, filespec, filetype, mode, access, lock,
            reclen, seg, offset, length, field
        ):
        """Open a file on a disk drive."""
        # parse the file spec to a definite native name
        if not self._native_root:
            # undefined disk drive: path not found
            raise error.BASICError(error.PATH_NOT_FOUND)
        # set default extension for programs
        if set(filetype).intersection(set(b'MPBA')):
            defext = b'BAS'
        else:
            defext = b''
        # translate the file name to something DOS-ish if necessary
        if mode == b'I':
            native_name = self._get_native_abspath(filespec, defext, isdir=False, create=False)
        else:
            # random files: try to open matching file
            # if it doesn't exist, use an all-caps 8.3 file name
            native_name = self._get_native_abspath(filespec, defext, isdir=False, create=True)
        # handle locks, open stream and create file object
        # don't open output or append files more than once
        # whether it's the same file is determined by DOS basename, i.e. excluding directories!
        dos_basename = self._get_dos_name_defext(filespec, defext)
        # obtain a lock
        self._locks.open_file(dos_basename, number, mode, lock, access)
        try:
            # open the underlying stream
            fhandle = self.open_stream(native_name, filetype, mode)
            # apply the BASIC file wrapper
            return self._create_file_object(
                fhandle, filetype, mode, number, field, reclen, seg, offset, length
            )
        except Exception:
            self._locks.close_file(number)
            raise

    def open_stream(self, native_name, filetype, mode):
        """Open a stream on disk by os-native name with BASIC mode and access level."""
        try:
            # create file if in RANDOM or APPEND mode and doesn't exist yet
            # OUTPUT mode files are created anyway since they're opened with wb
            if ((mode == b'A' or mode == b'R') and not os.path.exists(native_name)):
                io.open(native_name, 'wb').close()
            if mode == b'A':
                f = io.open(native_name, 'r+b')
                # APPEND mode is only valid for text files (which are seekable);
                # first cut off EOF byte, if any.
                try:
                    f.seek(-1, 2)
                    if f.read(1) == b'\x1a':
                        f.seek(-1, 1)
                        f.truncate()
                except EnvironmentError:
                    pass
                f.close()
            access_mode = ACCESS_MODES[mode]
            stream = io.open(native_name, access_mode + 'b')
            return stream
        except EnvironmentError as e:
            handle_oserror(e)
        except TypeError:
            # TypeError: stat() argument 1 must be encoded string without null bytes, not str
            # bad file number, which is what GW throws for open chr$(0)
            raise error.BASICError(error.BAD_FILE_NUMBER)

    def _get_native_reldir(self, dospath):
        """Return the native dir path for a given BASIC dir path, relative to the root."""
        if b'/' in dospath:
            # bad file number - this is what GW produces here
            raise error.BASICError(error.BAD_FILE_NUMBER)
        if not self._native_root:
            # this drive letter is not available (not mounted)
            raise error.BASICError(error.PATH_NOT_FOUND)
        # find starting directory
        if dospath and dospath[:1] == b'\\':
            # absolute path specified
            cwd = []
        else:
            cwd = self._native_cwd.split(os.sep)
        # parse internal .. and . and double slashes
        dospath = ntpath.normpath(dospath)
        # parse leading . and .. and double slashes in relative path
        # if at root, just drop leading dots (this is what GW-BASIC does at drive root)
        dospath_elements = dospath.split(b'\\')
        while dospath_elements and dospath_elements[0] in (b'', b'.', b'..'):
            if dospath_elements[0] == b'..':
                cwd = cwd[:-1]
            dospath_elements = dospath_elements[1:]
        # prepend drive root path to allow filename matching
        path = os.path.join(self._native_root, *cwd)
        root_len = len(self._native_root) + (self._native_root[-1:] != os.sep)
        # find the native matches for each step in the path
        for dos_elem in dospath_elements:
            # find a matching directory for every step in the path;
            native_elem = self._get_native_name(
                path, dos_elem, defext=b'', isdir=True, create=False
            )
            # append found name to path
            path = os.path.join(path, native_elem)
        # return relative path only
        return path[root_len:]

    def _get_native_abspath(self, path, defext, isdir, create):
        """\
            Find os-native path to match the given BASIC path; apply default extension.
            path: bytes             requested DOS path to file on this device
            defext: bytes           default extension, to apply if no dot in basename
            create: bool            allow creating a new name for the proposed DOS name
                                    if not, throw if matched basename does not exist
            isdir: bool             basename should refer to a directory
        """
        # substitute drives and cwds
        # always use Path Not Found error if not found at this stage
        dos_dirname, name = ntpath.split(path)
        native_relpath = self._get_native_reldir(dos_dirname)
        # return absolute path to file
        path = os.path.join(self._native_root, native_relpath)
        if name:
            path = os.path.join(path, self._get_native_name(path, name, defext, isdir, create))
        # get full normalised path
        return os.path.abspath(path)

    def chdir(self, dos_path):
        """Change working directory to given BASIC path."""
        # get drive path and relative path
        native_relpath = self._get_native_reldir(dos_path)
        # set cwd for the specified drive
        self._native_cwd = native_relpath

    def mkdir(self, dos_path):
        """Create directory at given BASIC path."""
        safe(os.mkdir, self._get_native_abspath(dos_path, defext=b'', isdir=True, create=True))

    def rmdir(self, dos_path):
        """Remove directory at given BASIC path."""
        safe(os.rmdir, self._get_native_abspath(dos_path, defext=b'', isdir=True, create=False))

    def kill(self, dos_pathmask):
        """Remove regular files that match given BASIC path and mask."""
        native_dir, _, dos_mask = self._split_pathmask(dos_pathmask)
        _, files = self._get_dirs_files(native_dir)
        # filter according to mask
        trunkmask, extmask = dos_splitext(dos_mask)
        dos_to_native = {
            self._get_dos_display_name(native_dir, _native_name): _native_name
            for _native_name in files
        }
        to_kill_dos = []
        for dos_name in dos_to_native:
            trunk, ext = dos_splitext(dos_name)
            if dos_name_matches(trunk, trunkmask) and dos_name_matches(ext, extmask):
                to_kill_dos.append(dos_name)
        to_kill = [
            # NOTE that this depends on display names NOT being legal names for overlong names
            # i.e. a + is included at the end of the display name which is not legal
            os.path.join(native_dir, dos_to_native[_dos_name])
            for _dos_name in to_kill_dos
            if (
                dos_is_legal_name(_dos_name) and
                not is_hidden(os.path.join(native_dir, dos_to_native[_dos_name]))
            )
        ]
        if not to_kill:
            raise error.BASICError(error.FILE_NOT_FOUND)
        for dos_path in to_kill_dos:
            # don't delete open files
            self.require_file_not_open(dos_path)
        for native_path in to_kill:
            safe(os.remove, native_path)

    def rename(self, old_dospath, new_dospath):
        """Rename a file or directory."""
        old_native_path = self._get_native_abspath(
            old_dospath, defext=b'', isdir=False, create=False
        )
        new_native_path = self._get_native_abspath(
            new_dospath, defext=b'', isdir=False, create=True
        )
        if os.path.exists(new_native_path):
            raise error.BASICError(error.FILE_ALREADY_EXISTS)
        safe(os.rename, old_native_path, new_native_path)

    def _split_pathmask(self, dos_pathmask):
        """Split pathmask into path and mask."""
        # forward slashes - file not found
        # GW-BASIC sometimes allows leading or trailing slashes
        # and then does weird things I don't understand.
        if b'/' in dos_pathmask:
            raise error.BASICError(error.FILE_NOT_FOUND)
        # note that ntpath would otherwise accept / as \\
        dos_path, dos_mask = ntpath.split(dos_pathmask)
        try:
            native_relpath = self._get_native_reldir(dos_path)
        except error.BASICError as e:
            # any path name problem in FILES: GW-BASIC throws file not found
            raise error.BASICError(error.FILE_NOT_FOUND)
        native_path = os.path.join(self._native_root, native_relpath)
        return native_path, native_relpath, dos_mask

    def _get_dirs_files(self, native_path):
        """Get native filenames for native path."""
        all_names = safe(os.listdir, native_path)
        dirs = [n for n in all_names if os.path.isdir(os.path.join(native_path, n))]
        fils = [n for n in all_names if not os.path.isdir(os.path.join(native_path, n))]
        return dirs, fils

    def listdir(self, pathmask):
        """Get directory listing."""
        native_path, _, dos_mask = self._split_pathmask(pathmask)
        fils = []
        if dos_mask in (b'.', b'..'):
            # following GW, we just show a single dot if asked for either . or ..
            dirs = [(b'', b'')]
        else:
            dirs, fils = self._get_dirs_files(native_path)
            # remove hidden files
            dirs = [d for d in dirs if not is_hidden(os.path.join(native_path, d))]
            fils = [f for f in fils if not is_hidden(os.path.join(native_path, f))]
            # filter according to mask
            dirs = self._filter_names(native_path, dirs + [u'.', u'..'], dos_mask)
            fils = self._filter_names(native_path, fils, dos_mask)
        # format contents
        return (
            [t.ljust(8) + (b'.' if e or not t else b' ') + e.ljust(3) + b'<DIR>' for t, e in dirs] +
            [t.ljust(8) + (b'.' if e or not t else b' ') + e.ljust(3) + b'     ' for t, e in fils]
        )

    def get_native_cwd(self):
        """Return the current working directory in native format."""
        return os.path.join(self._native_root, self._native_cwd)

    def get_cwd(self):
        """Return the current working directory in DOS format."""
        native_path = self._native_root
        dir_elems = []
        if self._native_cwd:
            for e in self._native_cwd.split(os.sep):
                dir_elems.append(self._get_dos_display_name(native_path, e))
                native_path += os.sep + e
        return self.letter + b':\\' + b'\\'.join(dir_elems)

    def get_free(self):
        """Return the number of free bytes on the drive."""
        return get_free_bytes(self._native_root)

    def require_file_exists(self, dospath):
        """Raise an error if the file is open or does not exist."""
        # this checks for existence with create=False
        self._get_native_abspath(dospath, defext=b'', isdir=False, create=False)

    def require_file_not_open(self, dos_basename):
        """Raise an error if the file is open."""
        if self._locks.list_open(dos_basename):
            raise error.BASICError(error.FILE_ALREADY_OPEN)

    ##########################################################################
    # DOS and native name conversion

    def _get_dos_name_defext(self, dos_name, defext):
        """Strip trailing whitepace and apply default extension to DOS name."""
        # ignore trailing whitespace
        dos_name = dos_name.rstrip()
        if defext and b'.' not in dos_name:
            dos_name += b'.' + defext
        return dos_name

    def _get_native_name(self, native_path, dos_name, defext, isdir, create):
        """Find or create a matching native file name for a given BASIC name."""
        # if the name contains a dot, do not apply the default extension
        # to maintain GW-BASIC compatibility, a trailing single dot matches the name
        # with no dots as well as the name with a single dot.
        # file names with more than one dot are not affected.
        # file spec         attempted matches
        # LongFileName      (1) LongFileName.BAS (2) LONGFILE.BAS
        # LongFileName.bas  (1) LongFileName.bas (2) LONGFILE.BAS
        # LongFileName.     (1) LongFileName. (2) LongFileName (3) LONGFILE
        # LongFileName..    (1) LongFileName.. (2) [does not try LONGFILE.. - not allowable]
        # Long.FileName.    (1) Long.FileName. (2) LONG.FIL
        #
        # don't accept leading whitespace (internal whitespace should be preserved)
        # note that DosBox removes internal whitespace, but MS-DOS does not
        name_err = error.PATH_NOT_FOUND if isdir else error.FILE_NOT_FOUND
        if dos_name != dos_name.lstrip():
            raise error.BASICError(name_err)
        dos_name = self._get_dos_name_defext(dos_name, defext)
        if dos_name[-1:] == b'.' and b'.' not in dos_name[:-1]:
            # ends in single dot; first try with dot
            # but if it doesn't exist, base everything off dotless name
            uni_name = self._codepage.bytes_to_unicode(dos_name, box_protect=False)
            if istype(native_path, uni_name, isdir):
                return uni_name
            dos_name = dos_name[:-1]
        # check if the name exists as-is; should also match Windows short names.
        uni_name = self._codepage.bytes_to_unicode(dos_name, box_protect=False)
        if istype(native_path, uni_name, isdir):
            return uni_name
        # original name does not exist; try matching dos-names or create one
        # normalise to 8.3
        norm_name = dos_normalise_name(dos_name)
        # check for non-legal characters & spaces (but clip off overlong names)
        if not dos_is_legal_name(norm_name):
            raise error.BASICError(error.BAD_FILE_NAME)
        fullname = dos_to_native_name(native_path, norm_name, isdir)
        if fullname:
            return fullname
        # not found
        if create:
            # create a new filename
            # we should have only ascii due to dos_is_legal_name check above
            return norm_name.decode('ascii', errors='replace')
        else:
            raise error.BASICError(name_err)

    def _get_dos_display_name(self, native_dirpath, native_name):
        """Convert native name to short name or (not normalised or even legal) dos-style name."""
        native_path = os.path.join(native_dirpath, native_name)
        # get the short name if it exists, keep long name otherwise
        native_path = get_short_pathname(native_path) or native_path
        native_name = os.path.basename(native_path)
        # see if we have a legal dos name that matches
        try:
            ascii_name = native_name.encode('ascii')
        except UnicodeEncodeError:
            pass
        else:
            if dos_is_legal_name(ascii_name):
                return dos_normalise_name(ascii_name)
        # convert to codepage
        cp_name = self._codepage.unicode_to_bytes(native_name, errors='replace')
        # clip overlong & mark as shortened
        trunk, ext = dos_splitext(cp_name)
        if len(trunk) > 8:
            trunk = trunk[:7] + b'+'
        if len(ext) > 3:
            ext = ext[:2] + b'+'
        return trunk + (b'.' if ext or not trunk else b'') + ext

    def _filter_names(self, native_dirpath, native_names, dos_mask):
        """Apply case-insensitive filename filter to display names."""
        dos_mask = dos_mask or b'*.*'
        trunkmask, extmask = dos_splitext(dos_mask)
        all_files = (self._get_dos_display_name(native_dirpath, name) for name in native_names)
        split = [dos_splitext(dos_name) for dos_name in all_files]
        return sorted(
            (trunk, ext) for (trunk, ext) in split
            if dos_name_matches(trunk, trunkmask) and dos_name_matches(ext, extmask)
        )


##############################################################################
# Native path utilities

def istype(native_path, native_name, isdir):
    """Return whether a file exists and is a directory or regular."""
    name = os.path.join(native_path, native_name)
    try:
        return os.path.isdir(name) if isdir else os.path.isfile(name)
    except (TypeError, ValueError):
        # name == u'\0' - python2 raises TypeError, python3 ValueError
        return False


##############################################################################
# Internal disk and bound files

@add_str
class BoundFile(object):
    """Bound internal file."""

    def __init__(self, device, codepage, file_name_or_object, name):
        """Initialise."""
        assert isinstance(name, bytes)
        self._device = device
        self._codepage = codepage
        self._file = file_name_or_object
        self._name = name

    def __enter__(self):
        """Context guard."""
        return self

    def __exit__(self, *dummies):
        """Context guard."""
        self._device.unbind(self._name)

    def get_stream(self, filetype, mode):
        """Get a native stream for the bound file."""
        try:
            if isinstance(self._file, (bytes, text_type)):
                return self._device.open_stream(self._file, filetype, mode)
            else:
                return self._file
        except EnvironmentError as e:
            handle_oserror(e)

    def __bytes__(self):
        """Get BASIC file name."""
        return b'%s:%s' % (self._device.letter, self._name)

    def __unicode__(self):
        """Get BASIC file name."""
        return self._codepage.bytes_to_unicode(bytes(self), box_protect=False)


@add_str
class NameWrapper(object):
    """Use normal file name as return value from bind_file."""

    def __init__(self, codepage, name):
        """Initialise."""
        if isinstance(name, text_type):
            name = codepage.unicode_to_bytes(name)
        self._file = name
        self._codepage = codepage

    def __enter__(self):
        """Context guard."""
        return self

    def __exit__(self, *dummies):
        """Context guard."""

    def __bytes__(self):
        """Get BASIC file name."""
        return self._file

    def __unicode__(self):
        """Get BASIC file name."""
        return self._codepage.bytes_to_unicode(bytes(self), box_protect=False)


class InternalDiskDevice(DiskDevice):
    """Internal disk device for special operations."""

    def __init__(self, letter, path, cwd, codepage, text_mode, soft_linefeed):
        """Initialise internal disk."""
        self._bound_files = {}
        DiskDevice.__init__(self, letter, path, cwd, codepage, text_mode, soft_linefeed)

    def bind(self, file_name_or_object, name=None):
        """Bind a native file name or object to an internal name."""
        if not name:
            # get unused 7-hexit string eg. #9ABCDEF
            try:
                name = random_id(7, prefix=b'#', exclude=self._bound_files)
            except RuntimeError: # pragma: no cover
                # unlikely
                logging.error('No free internal bound-file names available')
                raise error.BASICError(error.TOO_MANY_FILES)
        elif isinstance(name, text_type):
            name = self._codepage.unicode_to_bytes(name)
        self._bound_files[name] = BoundFile(self, self._codepage, file_name_or_object, name)
        return self._bound_files[name]

    def unbind(self, name):
        """Unbind bound file."""
        if isinstance(name, text_type):
            name = self._codepage.unicode_to_bytes(name)
        del self._bound_files[name]

    def open(
            self, number, filespec, filetype, mode, access, lock,
            reclen, seg, offset, length, field
        ):
        """Open a file on the internal disk drive."""
        if filespec in self._bound_files:
            fhandle = self._bound_files[filespec].get_stream(filetype, mode)
            try:
                return self._create_file_object(fhandle, filetype, mode)
            except EnvironmentError as e:
                raise
                handle_oserror(e)
        else:
            return DiskDevice.open(
                self, number, filespec, filetype, mode, access, lock,
                reclen, seg, offset, length, field
            )

    def _split_pathmask(self, pathmask):
        """Split pathmask into path and mask."""
        if self._native_root:
            return DiskDevice._split_pathmask(self, pathmask)
        else:
            return u'', u'', pathmask.upper() or b'*.*'

    def _get_dirs_files(self, path):
        """get native filenames for native path."""
        if self._native_root:
            dirs, files = DiskDevice._get_dirs_files(self, path)
        else:
            dirs, files = [], []
        files += self._bound_files.keys()
        return dirs, files

    def get_cwd(self):
        """Return the current working directory in DOS format."""
        if self._native_root:
            return DiskDevice.get_cwd(self)
        else:
            return self.letter + b':\\'

    def get_free(self):
        """Return the number of free bytes on the drive."""
        if self._native_root:
            return DiskDevice.get_free(self)
        else:
            return 0



"""
PC-BASIC - machine.py
Machine emulation and memory model

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import struct
import logging

from ..compat import iteritems, int2byte

from .data import NAME, VERSION, COPYRIGHT
from .base import error
from . import values
from . import devices
from .display import modes


# ROM copyright notice
NOTICE = bytearray(
    b'%s %s\r%s\r' % tuple(_.encode('ascii', 'ignore') for _ in (NAME, VERSION, COPYRIGHT))
)


###############################################################################

class MachinePorts(object):
    """Machine ports."""

    # time delay for port value to drop to 0 on maximum reading.
    #  use 100./255. for 100ms.
    joystick_time_factor = 75. / 255.

    def __init__(self, queues, values, display, keyboard, stick, files):
        """Initialise machine ports."""
        self._values = values
        self._queues = queues
        self._keyboard = keyboard
        self._stick = stick
        self._display = display
        # parallel port base address:
        # http://retired.beyondlogic.org/spp/parallel.htm
        # 3BCh - 3BFh  Used for Parallel Ports which were incorporated on to Video Cards
        #              - Doesn't support ECP addresses
        # 378h - 37Fh  Usual Address For LPT 1
        # 278h - 27Fh  Usual Address For LPT 2
        self.lpt_device = [files.get_device(b'LPT1:'), files.get_device(b'LPT2:')]
        # serial port base address:
        # http://www.petesqbsite.com/sections/tutorials/zines/qbnews/9-com_ports.txt
        #            COM1             &H3F8
        #            COM2             &H2F8
        #            COM3             &H3E8 (not implemented)
        #            COM4             &H2E8 (not implemented)
        self.com_base = {0x3f8: 0, 0x2f8: 1}
        self.com_device = [files.get_device(b'COM1:'), files.get_device(b'COM2:')]
        self.com_enable_baud_write = [False, False]
        self.com_baud_divisor = [0, 0]
        self.com_break = [False, False]

    def usr_(self, args):
        """USR: get value of machine-code function; not implemented."""
        num, = args
        logging.warning('USR function not implemented.')
        raise error.BASICError(error.IFC)

    def inp_(self, args):
        """INP: get value from machine port."""
        num, = args
        port = values.to_int(num, unsigned=True)
        inp = self.inp(port)
        # return as unsigned int
        if inp < 0:
            inp += 0x10000
        return self._values.new_integer().from_int(inp)

    def inp(self, port):
        """Get the value in an emulated machine port."""
        keyboard = self._keyboard
        stick = self._stick
        # keyboard
        if port == 0x60:
            return keyboard.last_scancode
        # game port (joystick)
        elif port == 0x201:
            value = (
                (not stick.is_firing[0][0]) * 0x40 +
                (not stick.is_firing[0][1]) * 0x20 +
                (not stick.is_firing[1][0]) * 0x10 +
                (not stick.is_firing[1][1]) * 0x80)
            decay = stick.decay()
            if decay < stick.axis[0][0] * self.joystick_time_factor:
                value += 0x04
            if decay < stick.axis[0][1] * self.joystick_time_factor:
                value += 0x02
            if decay < stick.axis[1][0] * self.joystick_time_factor:
                value += 0x01
            if decay < stick.axis[1][1] * self.joystick_time_factor:
                value += 0x08
            return value
        elif port in (0x379, 0x279):
            # parallel port input ports
            # http://www.aaroncake.net/electronics/qblpt.htm
            # http://retired.beyondlogic.org/spp/parallel.htm
            lpt_port_nr = 0 if port >= 0x378 else 1
            base_addr = {0: 0x378, 1: 0x278}
            if self.lpt_device[lpt_port_nr].stream is None:
                return 0
            # get status port
            busy, ack, paper, select, err = self.lpt_device[lpt_port_nr].stream.get_status()
            return busy * 0x80 | ack * 0x40 | paper * 0x20 | select * 0x10 | err * 0x8
        else:
            # serial port machine ports
            # http://www.qb64.net/wiki/index.php/Port_Access_Libraries#Serial_Communication_Registers
            # http://control.com/thread/1026221083
            for base_addr, com_port_nr in iteritems(self.com_base):
                com_port = self.com_device[com_port_nr]
                if not com_port.available():
                    continue
                # Line Control Register: base_address + 3 (r/w)
                if port == base_addr + 3:
                    _, parity, bytesize, stopbits = com_port.get_params()
                    value = self.com_enable_baud_write[com_port_nr] * 0x80
                    value += self.com_break[com_port_nr] * 0x40
                    value += {b'S': 0x38, b'M': 0x28, b'E': 0x18, b'O': 0x8, b'N': 0}[parity]
                    if stopbits > 1:
                        value += 0x4
                    value += bytesize - 5
                    return value
                # Line Status Register: base_address + 5 (read only)
                elif port == base_addr + 5:
                    # bit 6: data holding register empty
                    # bit 5: transmitter holding register empty
                    # distinction between bit 5 and 6 not implemented
                    # bit 0: data ready
                    # other bits not implemented:
                    #   1 - overrun, 2 - parity 3 - framing errors;
                    #   4 - break interrupt; 7 - at least one error in received FIFO
                    in_waiting, out_waiting = com_port.io_waiting()
                    return (1-out_waiting) * 0x60 + in_waiting
                # Modem Status Register: base_address + 6 (read only)
                elif port == base_addr + 6:
                    cd, ri, dsr, cts = com_port.get_pins()
                    # delta bits not implemented
                    return (cd*0x80 + ri*0x40 + dsr*0x20 + cts*0x10)
            # addr isn't one of the covered ports
            return 0

    def out_(self, args):
        """OUT: send a byte to a machine port."""
        addr = values.to_int(next(args), unsigned=True)
        val = values.to_int(next(args))
        error.range_check(0, 255, val)
        list(args)
        if addr == 0x201:
            # game port reset
            self._stick.reset_decay()
        elif addr == 0x3c5:
            # officially, requires OUT &H3C4, 2 first (not implemented)
            self._display.mode.memorymap.set_plane_mask(val)
        elif addr == 0x3cf:
            # officially, requires OUT &H3CE, 4 first (not implemented)
            self._display.mode.memorymap.set_plane(val)
        elif addr == 0x3d8:
            # CGA mode control register, see http://www.seasip.info/VintagePC/cga.html
            # bit 5 - enable blink (1) show blink as bright background (0) (not implemented)
            # bit 4 - select 640x200x2 mode (not implemented)
            # bit 3 - (1) enable video output (0) disable, show all as background (not implemented)
            # bit 2 - (1) disable colorburst (0) enable colorburst
            # bit 1 - (1) graphics mode (0) text mode (not implemented)
            # bit 0 - high resolution text (?) (not implemented)
            #OUT &H3D8,&H1A: REM enable color burst
            #OUT &H3D8,&H1E: REM disable color burst
            # 0x1a == 0001 1010     0x1e == 0001 1110
            self._display.colourmap.set_colorburst(val & 4 == 0)
        elif addr == 0x3d9:
            # CGA colour control register, see http://www.seasip.info/VintagePC/cga.html
            # bit 5 - palette 0 = r/g/y 1 = c/m/y/k (320x200x4 only)
            # bit 4 - 1 = high intensity 0 = low intensity (320x200x4 only)
            # bits 3-0: Border / Background / Foreground (not implemented)
            #    These 4 bits select one of the 16 CGA colours
            #    (bit 3 = Intensity, Bit 2 = Red, Bit 1 = Green, Bit 0 = Blue).
            #    In text modes, this colour is used for the border (overscan).
            #    In 320x200 graphics modes, it is used for the background and border.
            #    In 640x200 mode, it is used for the foreground colour.
            self._display.colourmap.set_cga4_palette(bool(val & 0x10))
            self._display.colourmap.set_cga4_intensity(bool(val & 0x8))
        elif addr in (0x378, 0x37A, 0x278, 0x27A):
            # parallel port output ports
            # http://www.aaroncake.net/electronics/qblpt.htm
            # http://retired.beyondlogic.org/spp/parallel.htm
            lpt_port_nr = 0 if addr >= 0x378 else 1
            base_addr = {0: 0x378, 1: 0x278}
            if self.lpt_device[lpt_port_nr].stream is None:
                return
            if addr - base_addr[lpt_port_nr] == 0:
                # set data port
                self.lpt_device[lpt_port_nr].stream.write(int2byte(val))
            else:
                # set control port
                self.lpt_device[lpt_port_nr].stream.set_control(
                    select=val & 0x8, init=val&0x4, lf=val&0x2, strobe=val&0x1
                )
        else:
            # serial port machine ports
            # http://www.qb64.net/wiki/index.php/Port_Access_Libraries#Serial_Communication_Registers
            # http://control.com/thread/1026221083
            for base_addr, com_port_nr in iteritems(self.com_base):
                com_port = self.com_device[com_port_nr]
                if not com_port.available():
                    continue
                # ports at base addr and the next one are used for writing baud rate
                # (among other things that aren't implemented)
                if addr in (base_addr, base_addr+1) and self.com_enable_baud_write[com_port_nr]:
                    if addr == base_addr:
                        self.com_baud_divisor[com_port_nr] = (
                            (self.com_baud_divisor[com_port_nr] & 0xff00) + val
                        )
                    elif addr == base_addr + 1:
                        self.com_baud_divisor[com_port_nr] = (
                            val*0x100 + (self.com_baud_divisor[com_port_nr] & 0xff)
                        )
                    if self.com_baud_divisor[com_port_nr]:
                        baudrate, parity, bytesize, stopbits = com_port.get_params()
                        baudrate = 115200 // self.com_baud_divisor[com_port_nr]
                        com_port.set_params(baudrate, parity, bytesize, stopbits)
                # Line Control Register: base_address + 3 (r/w)
                elif addr == base_addr + 3:
                    baudrate, parity, bytesize, stopbits = com_port.get_params()
                    if val & 0x80:
                        self.com_enable_baud_write[com_port_nr] = True
                    # break condition
                    self.com_break[com_port_nr] = (val & 0x40) != 0
                    # parity
                    parity = {0x38: b'S', 0x28: b'M', 0x18: b'E', 0x8: b'O', 0: b'N'}[val&0x38]
                    # stopbits
                    if val & 0x4:
                        # 2 or 1.5 stop bits
                        stopbits = 2
                    else:
                        # 1 stop bit
                        stopbits = 1
                    # set byte size to 5, 6, 7, 8
                    bytesize = (val & 0x3) + 5
                    com_port.set_params(baudrate, parity, bytesize, stopbits)
                    com_port.set_pins(brk=self.com_break[com_port_nr])
                # Modem Control Register: base_address + 4 (r/w)
                elif addr == base_addr + 4:
                    com_port.set_pins(rts=val & 0x2, dtr=val & 0x1)

    def wait_(self, args):
        """WAIT: wait for a machine port."""
        addr = values.to_int(next(args), unsigned=True)
        ander = values.to_int(next(args))
        error.range_check(0, 255, ander)
        xorer = next(args)
        if xorer is None:
            xorer = 0
        else:
            xorer = values.to_int(xorer)
        error.range_check(0, 255, xorer)
        list(args)
        while (self.inp(addr) ^ xorer) & ander == 0:
            self._queues.wait()


###############################################################################
# Memory

class Memory(object):
    """Memory model."""

    # lowest (EGA) video memory address; max 128k reserved for video
    video_segment = 0xa000
    # read only memory
    rom_segment = 0xf000
    # segment that holds ram font
    ram_font_segment = 0xc000

    # where to find the rom font (chars 0-127)
    rom_font_addr = 0xfa6e
    # where to find the ram font (chars 128-254)
    ram_font_addr = 0x500

    key_buffer_offset = 30
    blink_enabled = True

    def __init__(
            self, values, data_memory, files, display, keyboard,
            font_8, interpreter, peek_values, syntax
        ):
        """Initialise memory."""
        self._values = values
        # data segment initialised elsewhere
        self._memory = data_memory
        # device access needed for COM and LPT ports
        # files access for BLOAD and BSAVE
        self._files = files
        # screen access needed for video memory
        self._display = display
        # keyboard buffer access
        self.keyboard = keyboard
        # interpreter, for runmode check
        self.interpreter = interpreter
        # 8-pixel font
        self.font_8 = font_8
        # initial DEF SEG
        self.segment = self._memory.data_segment
        # pre-defined PEEK outputs
        self._peek_values = peek_values
        # tandy syntax
        self._syntax = syntax

    def peek_(self, args):
        """PEEK: Retrieve the value at an emulated memory location."""
        addr, = args
        # no peeking the program code (or anywhere) in protected mode
        if self._memory.program.protected and not self.interpreter.run_mode:
            raise error.BASICError(error.IFC)
        addr = values.to_int(addr, unsigned=True)
        addr += self.segment * 0x10
        return self._values.new_integer().from_int(self._get_memory(addr))

    def poke_(self, args):
        """POKE: Set the value at an emulated memory location."""
        addr = values.to_int(next(args), unsigned=True)
        if self._memory.program.protected and not self.interpreter.run_mode:
            raise error.BASICError(error.IFC)
        val, = args
        val = values.to_int(val)
        error.range_check(0, 255, val)
        if addr < 0:
            addr += 0x10000
        addr += self.segment * 0x10
        self._set_memory(addr, val)

    def bload_(self, args):
        """BLOAD: Load a file into a block of memory."""
        if self._memory.program.protected and not self.interpreter.run_mode:
            raise error.BASICError(error.IFC)
        name = values.next_string(args)
        offset = next(args)
        if offset is not None:
            offset = values.to_int(offset, unsigned=True)
        list(args)
        with self._files.open(0, name, filetype=b'M', mode=b'I') as g:
            # size gets ignored; even the \x1a at the end gets dumped onto the screen.
            seg = g.seg
            if offset is None:
                offset = g.offset
            buf = bytearray(g.read())
            # remove any EOF marker at end
            if buf and buf[-1] == 0x1a:
                buf = buf[:-1]
            # Tandys repeat the header at the end of the file
            if self._syntax == 'tandy':
                buf = buf[:-7]
            addr = seg * 0x10 + offset
            self._set_memory_block(addr, buf)

    def bsave_(self, args):
        """BSAVE: Save a block of memory into a file."""
        if self._memory.program.protected and not self.interpreter.run_mode:
            raise error.BASICError(error.IFC)
        name = values.next_string(args)
        offset = values.to_int(next(args), unsigned=True)
        length = values.to_int(next(args), unsigned=True)
        list(args)
        with self._files.open(
                    0, name, filetype=b'M', mode=b'O',
                    seg=self.segment, offset=offset, length=length
                ) as g:
            addr = self.segment * 0x10 + offset
            g.write(bytes(self._get_memory_block(addr, length)))
            # Tandys repeat the header at the end of the file
            if self._syntax == 'tandy':
                g.write(
                    devices.TYPE_TO_MAGIC[b'M'] + struct.pack('<HHH', self.segment, offset, length)
                )

    def def_seg_(self, args):
        """DEF SEG: Set segment."""
        segment, = args
        # &hb800: text screen buffer; &h13d: data segment
        if segment is None:
            self.segment = self._memory.data_segment
        else:
            # def_seg() accepts signed values
            self.segment = values.to_int(segment, unsigned=True)
            if self.segment < 0:
                self.segment += 0x10000

    def def_usr_(self, args):
        """DEF USR: Define machine language function."""
        usr, addr = args
        addr = values.to_integer(addr, unsigned=True)
        logging.warning('DEF USR statement not implemented')

    def call_(self, args):
        """CALL or CALLS: Call machine language procedure."""
        addr_var = next(args)
        # call procedure address must be numeric
        if self._memory.complete_name(addr_var)[-1:] == values.STR:
            # type mismatch
            raise error.BASICError(error.TYPE_MISMATCH)
        # ignore well-shaped arguments
        list(args)
        logging.warning('CALL/CALLS statement not implemented')

    ###########################################################################
    # IMPLEMENTATION

    def _get_memory(self, addr):
        """Retrieve the value at an emulated memory location."""
        try:
            # try if there's a preset value
            return self._peek_values[addr]
        except KeyError:
            if addr >= self.rom_segment*0x10:
                # ROM font
                return max(0, self._get_rom_memory(addr))
            elif addr >= self.ram_font_segment*0x10:
                # RAM font
                return max(0, self._get_font_memory(addr))
            elif addr >= self.video_segment*0x10:
                # graphics and text memory
                return max(0, self._get_video_memory(addr))
            elif addr >= self._memory.data_segment*0x10:
                return max(0, self._memory.get_memory(addr))
            elif addr >= 0:
                return max(0, self._get_low_memory(addr))
            else:
                return 0

    def _set_memory(self, addr, val):
        """Set the value at an emulated memory location."""
        if addr >= self.rom_segment*0x10:
            # ROM includes font memory
            pass
        elif addr >= self.ram_font_segment*0x10:
            # RAM font memory
            self._set_font_memory(addr, val)
        elif addr >= self.video_segment*0x10:
            # graphics and text memory
            self._set_video_memory(addr, val)
        elif addr >= self._memory.data_segment*0x10:
            self._memory.set_memory(addr, val)
        elif addr >= 0:
            self._set_low_memory(addr, val)

    def _get_memory_block(self, addr, length):
        """Retrieve a contiguous block of bytes from memory."""
        block = bytearray()
        if addr >= self.video_segment*0x10:
            video_len = 0x20000 - (addr - self.video_segment*0x10)
            # graphics and text memory - specialised call
            block += self._get_video_memory_block(addr, min(length, video_len))
            addr += video_len
            length -= video_len
        for a in range(addr, addr+length):
            block.append(max(0, self._get_memory(a)))
        return block

    def _set_memory_block(self, addr, buf):
        """Set a contiguous block of bytes in memory."""
        if addr >= self.video_segment*0x10:
            video_len = 0x20000 - (addr - self.video_segment*0x10)
            # graphics and text memory - specialised call
            self._set_video_memory_block(addr, buf[:video_len])
            addr += video_len
            buf = buf[video_len:]
        for a in range(len(buf)):
            self._set_memory(addr + a, buf[a])


    ###############################################################
    # video memory model

    def _get_video_memory(self, addr):
        """Retrieve a byte from video memory."""
        return self._display.mode.memorymap.get_memory(self._display, addr, 1)[0]

    def _set_video_memory(self, addr, val):
        """Set a byte in video memory."""
        with self._display.text_screen.collect_updates():
            self._display.mode.memorymap.set_memory(self._display, addr, [val])

    def _get_video_memory_block(self, addr, length):
        """Retrieve a contiguous block of bytes from video memory."""
        return bytearray(self._display.mode.memorymap.get_memory(self._display, addr, length))

    def _set_video_memory_block(self, addr, some_bytes):
        """Set a contiguous block of bytes in video memory."""
        with self._display.text_screen.collect_updates():
            self._display.mode.memorymap.set_memory(self._display, addr, some_bytes)

    ###############################################################################

    def _get_rom_memory(self, addr):
        """Retrieve data from ROM."""
        addr -= self.rom_segment*0x10
        if addr == 0xfffe:
            # machine ID byte
            # see http://stanislavs.org/helppc/id_bytes.html
            # FF	Original IBM PC  4/24/81
    		# FE	IBM XT (Original)
    		# FD	PCjr
    		# FC	IBM AT, XT 286, PS/1, PS/2 Model 50/60
    		# FB	IBM 256/640K XT (aka XT/2)
    		# FA	IBM PS/2 Model 30
    		# F9	IBM PC Convertible
    		# F8	IBM PS/2 Model 80/70
    		# B6    Hewlett Packard 110
    		# 9A	Compaq Plus
    		# 2D	Compaq PC
            # most clones including Tandy return FF (IBM PC) for compatibility
            # http://nerdlypleasures.blogspot.co.uk/2012/06/ibm-pcjr-and-tandy-1000-games.html
            if self._syntax == 'pcjr':
                return 0xfd
            return 0xff
        elif addr >= 0xe00e and addr < 0xe00e + 80:
            # version & copyright info instead of IBM BIOS copyright notice
            pos = addr - 0xe00e
            try:
                return NOTICE[pos]
            except IndexError:
                return -1
        else:
            # ROM font
            addr -= self.rom_font_addr
            char = addr // 8
            if char > 127 or char < 0:
                return -1
            return self.font_8.get_byte(char, addr%8)

    def _get_font_memory(self, addr):
        """Retrieve RAM font data."""
        addr -= self.ram_font_segment*0x10 + self.ram_font_addr
        char = addr // 8 + 128
        if char < 128 or char > 254:
            return -1
        return self.font_8.get_byte(char, addr%8)

    def _set_font_memory(self, addr, value):
        """Retrieve RAM font data."""
        addr -= self.ram_font_segment*0x10 + self.ram_font_addr
        char = addr // 8 + 128
        if char < 128 or char > 254:
            return
        self.font_8.set_byte(char, addr%8, value)

    #################################################################################


    def _get_low_memory(self, addr):
        """Retrieve data from low memory."""
        addr -= 0
        # from MEMORY.ABC: PEEKs and POKEs (Don Watkins)
        # http://www.qbasicnews.com/abc/showsnippet.php?filename=MEMORY.ABC&snippet=6
        # 108-115 control Ctrl-break capture; not implemented (see PC Mag POKEs)
        if addr == 124:
            return self.ram_font_addr % 256
        elif addr == 125:
            return self.ram_font_addr // 256
        elif addr == 126:
            return self.ram_font_segment % 256
        elif addr == 127:
            return self.ram_font_segment // 256
        # 1040 monitor type
        elif addr == 1040:
            if self._display.is_monochrome:
                # mono
                return 48 + 6
            else:
                # 80x25 graphics
                return 32 + 6
        # http://textfiles.com/programming/peekpoke.txt
        #   "(PEEK (1041) AND 14)/2" WILL PROVIDE NUMBER OF RS232 PORTS INSTALLED.
        #   "(PEEK (1041) AND 16)/16" WILL PROVIDE NUMBER OF GAME PORTS INSTALLED.
        #   "(PEEK (1041) AND 192)/64" WILL PROVIDE NUMBER OF PRINTERS INSTALLED.
        elif addr == 1041:
            return (
                2 * (
                    self._files.device_available(b'COM1:') +
                    self._files.device_available(b'COM2:')
                ) + 16 + 64 * (
                    self._files.device_available(b'LPT1:') +
                    self._files.device_available(b'LPT2:') +
                    self._files.device_available(b'LPT3:')
                )
            )
        # &h40:&h17 keyboard flag
        # &H80 - Insert state active
        # &H40 - CapsLock state has been toggled
        # &H20 - NumLock state has been toggled
        # &H10 - ScrollLock state has been toggled
        # &H08 - Alternate key depressed
        # &H04 - Control key depressed
        # &H02 - Left shift key depressed
        # &H01 - Right shift key depressed
        elif addr == 1047:
            return self.keyboard.mod
        # &h40:&h18 keyboard flag
        # &H80 - Insert key is depressed
        # &H40 - CapsLock key is depressed
        # &H20 - NumLock key is depressed
        # &H10 - ScrollLock key is depressed
        # &H08 - Suspend key has been toggled
        # not implemented: peek(1048)==4 if sysrq pressed, 0 otherwise
        elif addr == 1048:
            return 0
        elif addr == 1049:
            return int(self.keyboard.keypad_ascii or 0)%256
        elif addr == 1050:
            # keyboard ring buffer starts at n+1024; lowest 1054
            return (self.keyboard.buf.start*2 + self.key_buffer_offset) % 256
        elif addr == 1051:
            return (self.keyboard.buf.start*2 + self.key_buffer_offset) // 256
        elif addr == 1052:
            # ring buffer ends at n + 1023
            return (self.keyboard.buf.stop*2 + self.key_buffer_offset) % 256
        elif addr == 1053:
            return (self.keyboard.buf.stop*2 + self.key_buffer_offset) // 256
        elif addr in range(1024+self.key_buffer_offset, 1024+self.key_buffer_offset+32):
            index = (addr-1024-self.key_buffer_offset) // 2
            odd = (addr-1024-self.key_buffer_offset) % 2
            c, scan = self.keyboard.buf.ring_read(index)
            if odd:
                return scan or 0
            elif c == b'':
                return 0
            else:
                # however, arrow keys (all extended scancodes?) give 0xe0 instead of 0
                return ord(c[0:1])
        # 1097 screen mode number
        elif addr == 1097:
            # these are the low-level mode numbers used by mode switching interrupt
            return modes.get_mode_number(self._display.mode, self._display.colorswitch)
        # 1098, 1099 screen width
        elif addr == 1098:
            return self._display.mode.width % 256
        elif addr == 1099:
            return self._display.mode.width // 256
        # 1100, 1101 graphics page buffer size (32k for screen 9, 4k for screen 0)
        # 1102, 1103 zero (PCmag says graphics page buffer offset)
        elif addr == 1100:
            return self._display.mode.memorymap.page_size % 256
        elif addr == 1101:
            return self._display.mode.memorymap.page_size // 256
        # 1104 + 2*n (cursor column of page n) - 1
        # 1105 + 2*n (cursor row of page n) - 1
        # we only keep track of one row,col position
        elif addr in range(1104, 1120, 2):
            return self._display.text_screen.current_col - 1
        elif addr in range(1105, 1120, 2):
            return self._display.text_screen.current_row - 1
        # 1120, 1121 cursor shape
        elif addr == 1120:
            # to_line
            return self._display.cursor.shape[1]
        elif addr == 1121:
            # from_line
            return self._display.cursor.shape[0]
        # 1122 visual page number
        elif addr == 1122:
            return self._display.vpagenum
        # 1125 screen mode info
        elif addr == 1125:
            return self._display.get_mode_info_byte()
        # 1126 color
        elif addr == 1126:
            return self._display.get_colour_info_byte()
        # 1296, 1297: zero (PCmag says data segment address)
        return -1

    def _set_low_memory(self, addr, value):
        """Set data in low memory."""
        addr -= 0
        if addr == 1047:
            self.keyboard.mod = value
        # from basic_ref_3.pdf: the keyboard buffer may be cleared with
        # DEF SEG=0: POKE 1050, PEEK(1052)
        elif addr == 1050:
            # keyboard ring buffer starts at n+1024; lowest 1054
            self.keyboard.buf.ring_set_boundaries(
                (value - self.key_buffer_offset) // 2,
                self.keyboard.buf.stop
            )
        elif addr == 1052:
            # ring buffer ends at n + 1023
            self.keyboard.buf.ring_set_boundaries(
                self.keyboard.buf.start,
                (value - self.key_buffer_offset) // 2
            )
        elif addr in range(1024+self.key_buffer_offset, 1024+self.key_buffer_offset+32):
            index = (addr-1024-self.key_buffer_offset) // 2
            odd = (addr-1024-self.key_buffer_offset) % 2
            c, scan = self.keyboard.buf.ring_read(index)
            if odd:
                scan = value
            elif value in (0, 0xe0):
                c = b''
            else:
                c = int2byte(value)
            self.keyboard.buf.ring_write(index, c, scan)



"""
PC-BASIC - interpreter.py
BASIC interpreter

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import struct

from .base import error
from .base import tokens as tk
from .base.tokens import DIGITS
from .base import codestream
from . import values


class Interpreter(object):
    """BASIC interpreter."""

    def __init__(
            self, queues, console, cursor, files, sound,
            values, memory, program, parser, basic_events
        ):
        """Initialise interpreter."""
        self._queues = queues
        self._basic_events = basic_events
        self._values = values
        self._memory = memory
        self._scalars = memory.scalars
        self._console = console
        self._cursor = cursor
        self._files = files
        self._sound = sound
        # program buffer
        self._program = program
        self._program_code = program.bytecode
        # direct line buffer
        self.direct_line = codestream.TokenisedStream()
        self.current_statement = 0
        # statement syntax parser
        self.parser = parser
        # line number tracing
        self.tron = False
        # pointer position: False for direct line, True for program
        self.run_mode = False
        # clear stacks
        self.clear_stacks_and_pointers()
        self._init_error_trapping()
        self.error_num = 0
        self.error_pos = 0
        self.set_pointer(False, 0)
        # interpreter is waiting for INPUT or LINE INPUT
        self.input_mode = False
        # interpreter is executing a command (needs console)
        self.set_parse_mode(False)
        # additional operations on program step (debugging)
        self.step = lambda token: None

    def __getstate__(self):
        """Pickle."""
        pickle_dict = self.__dict__.copy()
        # functions can't be pickled
        pickle_dict['step'] = None
        return pickle_dict

    def __setstate__(self, pickle_dict):
        """Unpickle."""
        self.__dict__.update(pickle_dict)
        self.step = lambda token: None
        # go to start of statement on resume
        ins = self.get_codestream()
        ins.seek(self.current_statement)
        if not self.parser.redo_on_break:
            # parse line number or : at start of statement
            if ins.read(1) in tk.END_LINE:
                # line number marker, new statement
                token = ins.read(4)
            ins.skip_to(tk.END_STATEMENT)

    def _init_error_trapping(self):
        """Initialise error trapping."""
        # True if error handling in progress
        self.error_handle_mode = False
        # statement pointer, run mode of error for RESUME
        self.error_resume = None
        # pointer to error trap
        self.on_error = None

    def parse(self):
        """Parse from the current pointer in current codestream."""
        while True:
            # update what basic events need to be handled
            self._queues.set_basic_event_handlers(self._basic_events.enabled)
            # check input and BASIC events. may raise Break, Reset or Exit
            self._queues.check_events()
            try:
                self.handle_basic_events()
                ins = self.get_codestream()
                self.current_statement = ins.tell()
                c = ins.skip_blank_read()
                # parse line number or : at start of statement
                if c in tk.END_LINE:
                    # line number marker, new statement
                    token = ins.read(4)
                    # end of program or truncated file
                    if token[:2] == b'\0\0' or len(token) < 4:
                        if c == b'\0' and self.error_resume:
                            # unfinished error handler: no RESUME (don't trap this)
                            self.error_handle_mode = True
                            # get line number right
                            raise error.BASICError(error.NO_RESUME, ins.tell()-len(token)-2)
                        # stream has ended
                        self.set_pointer(False)
                        return
                    if self.tron:
                        linenum = struct.unpack_from('<H', token, 2)
                        self._console.write(b'[%i]' % linenum)
                    self.step(token)
                elif c not in (b':', tk.THEN, tk.ELSE, tk.GOTO):
                    # new statement or branch of an IF statement allowed, nothing else
                    raise error.BASICError(error.STX)
                self.parser.parse_statement(ins)
            except error.BASICError as e:
                self.trap_error(e)

    def loop(self):
        """Run commands until control returns to user."""
        if not self.parse_mode:
            return
        try:
            # parse until break or end
            self.parse()
        except error.Break as e:
            self._sound.stop_all_sound()
            self._handle_break(e)
        # move pointer to the start of direct line (for both on and off!)
        self.set_pointer(False, 0)
        # return control to user
        self.set_parse_mode(False)

    def set_parse_mode(self, on):
        """Enter or exit parse mode."""
        self.parse_mode = on
        self._cursor.set_direct(not on)

    def _handle_break(self, e):
        """Handle a Break event."""
        # print ^C at current position
        if not self.input_mode and not e.stop:
            self._console.write(b'^C')
        # if we're in a program, save pointer
        pos = -1
        if self.run_mode:
            pos = self.current_statement
            if self.parser.redo_on_break:
                self.stop_pos = pos
            else:
                self._program.bytecode.skip_to(tk.END_STATEMENT)
                self.stop_pos = self._program.bytecode.tell()
        self.parser.redo_on_break = False
        if self.error_handle_mode:
            e.trapped_error_num = self.error_num
            # when the error is trapped, error position shifts to start of line
            line = self._program.get_line_number(self.error_pos)
            if 0 <= line <= 65535:
                e.trapped_error_pos = self._program.line_numbers[line]
            else:
                e.trapped_error_pos = self.error_pos
            #self.error_handle_mode = False
        # ensure we can handle the break like an error
        e.err = 0
        e.pos = pos
        raise e

    ###########################################################################
    # clear state

    def clear(self):
        """Clear all to be cleared for CLEAR statement."""
        # clear last error number (ERR) and line number (ERL)
        self.error_num, self.error_pos = 0, 0
        # disable error trapping
        self._init_error_trapping()
        # disable all event trapping (resets PEN to OFF too)
        self._basic_events.reset()
        # CLEAR also dumps for_next and while_wend stacks
        self.for_stack = []
        self.while_stack = []
        # reset stop/cont
        self.stop_pos = None
        # reset the DATA pointer
        self.data_pos = 0

    def clear_stacks_and_pointers(self):
        """Initialise the stacks and pointers for a new program."""
        # reset loop stacks
        self._clear_stacks()
        # reset program pointer
        self._program_code.seek(0)
        # reset stop/cont
        self.stop_pos = None
        # reset data reader
        self.data_pos = 0

    def _clear_stacks(self):
        """Clear loop and jump stacks."""
        # stop running if we were
        self.set_pointer(False)
        self.gosub_stack = []
        self.for_stack = []
        self.while_stack = []


    ###########################################################################
    # event and error handling

    def handle_basic_events(self):
        """Jump to user-defined event subs if events triggered."""
        if self._basic_events.suspend_all or not self.run_mode:
            return
        for event in self._basic_events.enabled:
            if (event.triggered and not event.stopped and event.gosub is not None):
                # release trigger
                event.triggered = False
                # stop this event while handling it
                event.stopped = True
                # execute 'ON ... GOSUB' subroutine;
                # attach handler to allow un-stopping event on RETURN
                self.jump_sub(event.gosub, event)

    def trap_error(self, e):
        """Handle a BASIC error through trapping."""
        if e.pos is None:
            if self.run_mode:
                e.pos = self._program_code.tell()-1
            else:
                e.pos = -1
        self.error_num = e.err
        self.error_pos = e.pos
        # don't jump if we're already busy handling an error
        if self.on_error is not None and self.on_error != 0 and not self.error_handle_mode:
            self.error_resume = self.current_statement, self.run_mode
            self.jump(self.on_error)
            self.error_handle_mode = True
            self._basic_events.suspend_all = True
        else:
            self.error_handle_mode = False
            self.set_pointer(False)
            raise e

    def erl_(self, args):
        """ERL: get line number of last error."""
        list(args)
        if self.error_pos == 0:
            pos = 0
        elif self.error_pos == -1:
            pos = 65535
        else:
            pos = self._program.get_line_number(self.error_pos)
        return self._values.new_single().from_int(pos)

    def err_(self, args):
        """ERR: get error code of last error."""
        list(args)
        return self._values.new_integer().from_int(self.error_num)

    ###########################################################################
    # jumps

    def set_pointer(self, new_runmode, pos=None):
        """Set program pointer to the given codestream and position."""
        # flush lpt1 on entering interactive mode
        if self.run_mode and not new_runmode:
            self._files.lpt1_file.do_print()
        self.run_mode = new_runmode
        # events are active in run mode
        if new_runmode:
            self._queues.set_basic_event_handlers(self._basic_events.enabled)
        else:
            self._queues.set_basic_event_handlers([])
        # keep the sound engine on to avoid delays in run mode
        self._sound.persist(new_runmode)
        # suppress cassette messages in run mode
        self._files.get_device(b'CAS1:').quiet(new_runmode)
        codestream = self.get_codestream()
        if pos is not None:
            # jump to position, if given
            codestream.seek(pos)
        else:
            # position at end - don't execute anything unless we jump
            codestream.seek(0, 2)

    def get_codestream(self):
        """Get the current codestream."""
        return self._program_code if self.run_mode else self.direct_line

    def jump(self, jumpnum, err=error.UNDEFINED_LINE_NUMBER):
        """Execute jump for a GOTO or RUN instruction."""
        if jumpnum is None:
            self.set_pointer(True, 0)
        else:
            try:
                # jump to target
                self.set_pointer(True, self._program.line_numbers[jumpnum])
            except KeyError:
                raise error.BASICError(err)

    def jump_sub(self, jumpnum, handler=None):
        """Execute jump for a GOSUB."""
        # set return position
        pos = self.get_codestream().tell()
        # record run mode before executing self.jump, as that sets the runmode to True
        run_mode = self.run_mode
        self.jump(jumpnum)
        self.gosub_stack.append((pos, run_mode, handler))

    def goto_(self, args):
        """GOTO: jump to line number."""
        self.jump(*args)

    def gosub_(self, args):
        """GOSUB: jump to subroutine."""
        self.jump_sub(*args)

    def return_(self, args):
        """Execute jump for a RETURN."""
        jumpnum, = args
        try:
            pos, orig_runmode, handler = self.gosub_stack.pop()
        except IndexError:
            raise error.BASICError(error.RETURN_WITHOUT_GOSUB)
        # returning from ON (event) GOSUB, re-enable event
        if handler:
            # if stopped explicitly using STOP, we wouldn't have got here;
            # if STOP is run inside the trap, no effect. OFF in trap: event off.
            handler.stopped = False
        if jumpnum is None:
            # go back to position of GOSUB
            self.set_pointer(orig_runmode, pos)
            # ignore rest of statement ('GOSUB 100 LAH' works just fine..)
            # but NOT if we jumped for an event, as we might have jumped from anywhere!
            if not handler:
                self.get_codestream().skip_to(tk.END_STATEMENT)
        else:
            # jump to specified line number
            self.jump(jumpnum)

    ###########################################################################
    # branches

    def if_(self, args):
        """IF: branching statement."""
        # get condition
        # avoid overflow: don't use bools.
        then_branch = not values.to_single(next(args)).is_zero()
        # cofunction only parses the branch we need
        # cofunction checks for line number
        branch = args.send(then_branch)
        # and completes
        list(args)
        # we may have a line number immediately after THEN or ELSE
        if branch is not None:
            self.jump(branch)
        # note that any :ELSE block encountered will be ignored automatically
        # since standalone ELSE is a no-op to end of line

    def on_jump_(self, args):
        """ON GOTO/GOSUB: calculated jump."""
        onvar = values.to_int(next(args))
        error.range_check(0, 255, onvar)
        jump_type = next(args)
        # only parse jumps (and errors!) up to our choice
        i = -1
        for i, jumpnum in enumerate(args):
            if i == onvar-1:
                # we counted the right number of commas, then didn't find a line number
                if jumpnum is None:
                    raise error.BASICError(error.STX)
                if jump_type == tk.GOTO:
                    self.jump(jumpnum)
                elif jump_type == tk.GOSUB:
                    self.jump_sub(jumpnum)
                return

    ###########################################################################
    # loops

    def for_(self, args):
        """Initialise a FOR loop."""
        # read variable
        varname = self._memory.complete_name(next(args))
        vartype = varname[-1:]
        start = values.to_type(vartype, next(args)).clone()
        # only raised after the TO has been parsed
        if vartype in (values.STR, values.DBL):
            raise error.BASICError(error.TYPE_MISMATCH)
        stop = values.to_type(vartype, next(args)).clone()
        step = next(args)
        if step is not None:
            step = values.to_type(vartype, step).clone()
        list(args)
        if step is None:
            # convert 1 to vartype
            step = self._values.from_value(1, varname[-1:])
        ins = self.get_codestream()
        # find NEXT
        forpos, nextpos = self._find_next(ins, varname)
        # initialise loop variable
        self._scalars.set(varname, start)
        # obtain a view of the loop variable
        self.for_stack.append((varname, stop, step, step.sign(), forpos, nextpos,))
        # empty loop: jump to NEXT without executing block
        if (start.gt(stop) if step.sign() >= 0 else stop.gt(start)):
            ins.seek(nextpos)
            self.iterate_loop()

    def _find_next(self, ins, varname):
        """Helper function for FOR: find matching NEXT."""
        endforpos = ins.tell()
        ins.skip_block(tk.FOR, tk.NEXT, allow_comma=True)
        if ins.skip_blank() not in (tk.NEXT, b','):
            # FOR without NEXT marked with FOR line number
            ins.seek(endforpos)
            raise error.BASICError(error.FOR_WITHOUT_NEXT)
        comma = (ins.read(1) == b',')
        # check var name for NEXT
        # no-var only allowed in standalone NEXT
        if ins.skip_blank() not in tk.END_STATEMENT:
            varname2 = self._memory.complete_name(self.parser.parse_name(ins))
        else:
            varname2 = None
        # get position and line number just after the matching variable in NEXT
        nextpos = ins.tell()
        if (comma or varname2) and varname2 != varname:
            # NEXT without FOR marked with NEXT line number, while we're only at FOR
            raise error.BASICError(error.NEXT_WITHOUT_FOR)
        ins.seek(endforpos)
        return endforpos, nextpos

    def next_(self, args):
        """Iterate a loop (NEXT)."""
        for varname in args:
            # increment counter, check condition
            if self.iterate_loop(varname):
                break

    def iterate_loop(self, varname=None):
        """Iterate a loop (NEXT)."""
        ins = self.get_codestream()
        # record the location after the variable
        pos = ins.tell()
        # find the matching NEXT record
        num = len(self.for_stack)
        for depth in range(num):
            varname2, stop, step, sgn, forpos, nextpos = self.for_stack[-depth-1]
            if pos == nextpos:
                if varname is not None and varname2 != self._memory.complete_name(varname):
                    # check once more for matches
                    # it has been checked at FOR, but DEFtypes may have changed.
                    raise error.BASICError(error.NEXT_WITHOUT_FOR)
                # only drop NEXT record if we've found a matching one
                self.for_stack = self.for_stack[:len(self.for_stack)-depth]
                break
        else:
            raise error.BASICError(error.NEXT_WITHOUT_FOR)
        # increment counter
        counter_view = self._scalars.view(varname2)
        counter_view.iadd(step)
        # check condition
        loop_ends = counter_view.gt(stop) if sgn > 0 else stop.gt(counter_view)
        if loop_ends:
            self.for_stack.pop()
        else:
            ins.seek(forpos)
        return not loop_ends

    def while_(self, args):
        """WHILE: enter while-loop."""
        list(args)
        ins = self.get_codestream()
        # find matching WEND
        whilepos, wendpos = self._find_wend(ins)
        self.while_stack.append((whilepos, wendpos))
        self._check_while_condition(ins, whilepos)

    def _find_wend(self, ins):
        """Helper function for WHILE: find matching WEND."""
        # just after WHILE token
        whilepos = ins.tell()
        ins.skip_block(tk.WHILE, tk.WEND)
        if ins.read(1) != tk.WEND:
            # WHILE without WEND
            ins.seek(whilepos)
            raise error.BASICError(error.WHILE_WITHOUT_WEND)
        ins.skip_to(tk.END_STATEMENT)
        wendpos = ins.tell()
        ins.seek(whilepos)
        return whilepos, wendpos

    def _check_while_condition(self, ins, whilepos):
        """Check condition of while-loop."""
        ins.seek(whilepos)
        # WHILE condition is zero?
        if not values.pass_number(self.parser.parse_expression(ins)).is_zero():
            # statement start is before WHILE token
            self.current_statement = whilepos-2
            ins.require_end()
        else:
            # ignore rest of line and jump to WEND
            _, wendpos = self.while_stack.pop()
            ins.seek(wendpos)

    def wend_(self, args):
        """WEND: iterate while-loop."""
        list(args)
        ins = self.get_codestream()
        pos = ins.tell()
        while True:
            if not self.while_stack:
                # WEND without WHILE
                raise error.BASICError(error.WEND_WITHOUT_WHILE)
            whilepos, wendpos = self.while_stack[-1]
            if pos == wendpos:
                break
            # not the expected WEND, we must have jumped out
            self.while_stack.pop()
        self._check_while_condition(ins, whilepos)

    ###########################################################################
    # DATA utilities

    def restore_(self, args):
        """Reset data pointer (RESTORE) """
        datanum = next(args)
        if datanum is None:
            self.data_pos = 0
        else:
            try:
                self.data_pos = self._program.line_numbers[datanum]
            except KeyError:
                raise error.BASICError(error.UNDEFINED_LINE_NUMBER)
        list(args)

    def read_(self, args):
        """READ: read values from DATA statement."""
        data_error = False
        for name, indices in args:
            name = self._memory.complete_name(name)
            current = self._program_code.tell()
            self._program_code.seek(self.data_pos)
            if self._program_code.peek() in tk.END_STATEMENT:
                # initialise - find first DATA
                self._program_code.skip_to_token(tk.DATA,)
            if self._program_code.read(1) not in (tk.DATA, b','):
                self._program_code.seek(current)
                raise error.BASICError(error.OUT_OF_DATA)
            self._program_code.skip_blank()
            if name[-1:] == values.STR:
                # for unquoted strings, payload starts at the first non-empty character
                address = self._program_code.tell_address()
                word = self._program_code.read_to((b',', b'"',) + tk.END_STATEMENT)
                if self._program_code.peek() == b'"':
                    if word == b'':
                        # nothing before the quotes, so this is a quoted string literal
                        # string payload starts after quote
                        address = self._program_code.tell_address() + 1
                        word = self._program_code.read_string().strip(b'"')
                    else:
                        # complete unquoted string literal
                        word += self._program_code.read_string()
                    if (self._program_code.skip_blank() not in (tk.END_STATEMENT + (b',',))):
                        raise error.BASICError(error.STX)
                else:
                    word = word.strip(self._program_code.blanks)
                value = self._values.from_str_at(word, address)
            else:
                word = self._program_code.read_number()
                value = self._values.from_repr(word, allow_nonnum=False)
                # anything after the number is a syntax error, but assignment has taken place)
                if (self._program_code.skip_blank() not in (tk.END_STATEMENT + (b',',))):
                    data_error = True
            # restore to current program location
            # to ensure any other errors in set_variable get the correct line number
            data_pos = self._program_code.tell()
            self._program_code.seek(current)
            self._memory.set_variable(name, indices, value=value)
            if data_error:
                self._program_code.seek(self.data_pos)
                raise error.BASICError(error.STX)
            else:
                self.data_pos = data_pos

    ###########################################################################
    # COMMON

    def gather_commons(self):
        """Get all COMMON declarations."""
        common_scalars = set()
        common_arrays = set()
        current = self._program_code.tell()
        self._program_code.seek(0)
        while self._program_code.skip_to_token(tk.COMMON):
            self._program_code.read(len(tk.COMMON))
            self._add_common_vars(common_scalars, common_arrays)
        self._program_code.seek(current)
        return common_scalars, common_arrays

    def _parse_common_args(self, ins):
        """Parse COMMON syntax."""
        if ins.skip_blank() in tk.END_STATEMENT:
            return
        while True:
            name = self.parser.parse_name(ins)
            bracket = ins.skip_blank_read_if((b'(', b'['))
            if bracket:
                # a literal is allowed but ignored;
                # for sqare brackets, it's a syntax error if omitted
                if (bracket == b'[') or ins.peek() in set(DIGITS) | set(tk.NUMBER):
                    x = self.parser.expression_parser.read_number_literal(ins)
                ins.require_read((b')', b']'))
            # entries with square brackets are completely ignored!
            if bracket != b'[':
                yield name, bracket
            if not ins.skip_blank_read_if((b',',)):
                break
        ins.require_end()

    def _add_common_vars(self, common_scalars, common_arrays):
        """COMMON: define variables to be preserved on CHAIN."""
        common_vars = list(self._parse_common_args(self._program_code))
        common_scalars.update(
            self._memory.complete_name(name)
            for name, brackets in common_vars if not brackets
        )
        common_arrays.update(
            self._memory.complete_name(name)
            for name, brackets in common_vars if brackets
        )

    ###########################################################################
    # callbacks

    def error_(self, args):
        """ERROR: simulate an error condition."""
        errn, = args
        errn = values.to_int(errn)
        error.range_check(1, 255, errn)
        raise error.BASICError(errn)

    def stop_(self, args):
        """STOP: break program execution and return to interpreter."""
        list(args)
        raise error.Break(stop=True)

    def cont_(self, args):
        """CONT: continue STOPped or ENDed execution."""
        list(args)
        if self.stop_pos is None:
            raise error.BASICError(error.CANT_CONTINUE)
        else:
            self.set_pointer(True, self.stop_pos)
        # IN GW-BASIC, weird things happen if you do GOSUB nn :PRINT "x"
        # and there's a STOP in the subroutine.
        # CONT then continues and the rest of the original line is executed, printing x
        # However, CONT:PRINT triggers a bug
        #  - a syntax error in a nonexistant line number is reported.
        # CONT:PRINT "y" results in neither x nor y being printed.
        # if a command is executed before CONT, x is not printed.
        # It would appear that GW-BASIC only partially overwrites the line buffer and
        # then jumps back to the original return location!
        # in this implementation, the CONT command will fully overwrite the line buffer
        # so x is not printed.

    def tron_(self, args):
        """TRON: trace on."""
        list(args)
        self.tron = True

    def troff_(self, args):
        """TROFF: trace off."""
        list(args)
        self.tron = False

    def on_error_goto_(self, args):
        """ON ERROR GOTO: define error trapping routine."""
        linenum, = args
        if linenum != 0 and linenum not in self._program.line_numbers:
            raise error.BASICError(error.UNDEFINED_LINE_NUMBER)
        self.on_error = linenum
        # pause soft-handling math errors so that we can catch them
        self._values.error_handler.suspend(linenum != 0)
        # ON ERROR GOTO 0 in error handler
        if self.on_error == 0 and self.error_handle_mode:
            # re-raise the error so that execution stops
            raise error.BASICError(self.error_num, self.error_pos)

    def resume_(self, args):
        """RESUME: resume program flow after error-trap."""
        if self.error_resume is None:
            # unset error handler
            self.on_error = 0
            raise error.BASICError(error.RESUME_WITHOUT_ERROR)
        # parse arguments
        where, = args
        start_statement, runmode = self.error_resume
        self.error_num = 0
        self.error_handle_mode = False
        self.error_resume = None
        self._basic_events.suspend_all = False
        if not where:
            # RESUME or RESUME 0
            self.set_pointer(runmode, start_statement)
        elif where == tk.NEXT:
            # RESUME NEXT
            self.set_pointer(runmode, start_statement)
            self.get_codestream().skip_to(tk.END_STATEMENT, break_on_first_char=False)
        else:
            # RESUME n
            self.jump(where)

    def def_fn_(self, args):
        """DEF FN: define a function."""
        fnname, = args
        fnname = self._memory.complete_name(fnname)
        # don't allow DEF FN in direct mode, as we point to the code in the stored program
        # this is raised before further syntax errors
        if not self.run_mode:
            raise error.BASICError(error.ILLEGAL_DIRECT)
        # arguments and expression are being read and parsed by UserFunctionManager
        self.parser.user_functions.define(fnname, self._program_code)

    def llist_(self, args):
        """LLIST: output program lines to LPT1: """
        line_range, = args
        for l in self._program.list_lines(*line_range):
            self._files.lpt1_file.write_line(l)
        # return to direct mode
        self.set_pointer(False)

    def renum_(self, args):
        """RENUM: renumber program line numbers."""
        new, old, step = args
        new, old = self._program.explicit_lines(new, old)
        if step is not None and step < 1:
            raise error.BASICError(error.IFC)
        old_to_new = self._program.renum(self._console, new, old, step)
        # stop running if we were
        # reset loop stacks
        self._clear_stacks()
        # renumber error handler
        if self.on_error:
            self.on_error = old_to_new[self.on_error]
        # renumber event traps
        for handler in self._basic_events.all:
            if handler.gosub:
                handler.set_jump(old_to_new[handler.gosub])



import sys
import io
import os
import codecs
import tempfile
import logging
from contextlib import contextmanager

from .base import PY2


class StreamWrapper(object):
    """Base class for delegated stream wrappers."""

    def __init__(self, stream):
        """Set up codec."""
        self._stream = stream

    def __getattr__(self, attr):
        return getattr(self._stream, attr)

    def __getstate__(self):
        return vars(self)

    def __setstate__(self, stdict):
        return vars(self).update(stdict)



def _open_named_devnull(name, mode):
    """Open a stream to /dev/null with the given stream name."""
    stream = StreamWrapper(open(os.devnull, mode))
    stream.name = name
    return stream


def fix_stdio():
    """Make sure sys.std* exist and are forgiving of errors."""
    # stdio may become None in GUI mode
    # fix them to devnull to ensure any i/o doesn't lead to crashes
    if not sys.stdin:
        sys.stdin = _open_named_devnull('<stdin>', 'r')
    if not sys.stdout:
        sys.stdout = _open_named_devnull('<stdout>', 'w')
    if not sys.stderr:
        sys.stderr = _open_named_devnull('<stderr>', 'w')

    if not PY2:
        # avoid UnicodeDecodeErrors when writing to terminal which doesn't support all of Unicode
        # e.g latin-1 locales or unsupported locales defaulting to ascii
        try:
            # this needs Python >= 3.7
            sys.stdout.reconfigure(errors='replace')
        except AttributeError:
            sys.stdout.__init__(sys.stdout.buffer, encoding=sys.stdout.encoding, errors='replace')
        try:
            sys.stderr.reconfigure(errors='replace')
        except AttributeError:
            sys.stderr.__init__(sys.stderr.buffer, encoding=sys.stderr.encoding, errors='replace')


if PY2:
    def is_writable_text_stream(stream):
        """Stream is a writable stream that expects unicode."""
        return isinstance(stream, (
            io.TextIOWrapper, io.StringIO,
            codecs.StreamReaderWriter, codecs.StreamWriter,
        ))

    def is_readable_text_stream(stream):
        """Stream is a readable stream that produces unicode."""
        return isinstance(stream, (
            io.TextIOWrapper, io.StringIO,
            codecs.StreamReaderWriter, codecs.StreamReader,
        ))
else:
    def is_writable_text_stream(stream):
        """Stream is a writable stream that expects unicode."""
        try:
            stream.write(u'')
        except TypeError:
            return False
        return True

    def is_readable_text_stream(stream):
        """Stream is a readable stream that produces unicode."""
        return isinstance(stream.read(0), type(u''))



# pause/quiet standard streams
# ----------------------------
# previously we had a version that redirected c-level streams, thereby catching non-python output
# based on https://eli.thegreenplace.net/2015/redirecting-all-kinds-of-stdout-in-python/
# or http://stackoverflow.com/questions/977840/redirecting-fortran-called-via-f2py-output-in-python/978264#978264
# however, that method either:
# - closes stdout/stderr, as in the above link; which means that any cached reference
#   to sys.stderr or sys.stdout now refers to a closed file, and leads to an IOError if flushed.
# - or doesn't close stdout/stderr, in which case the windows version of python 3 keeps a handle
#   to the old stdout, whcih is now an invalid handle, leading to a WindowsError when flushed.
#   background here:
#   - https://stackoverflow.com/questions/52373180/python-on-windows-handle-invalid-when-redirecting-stdout-writing-to-file
#   - https://stackoverflow.com/questions/902967/what-is-a-windows-handle/902969#902969
# it would be possible to set PYTHONLEGACYWINDOWSSTDIO and use the python2 code to deal with unicode output, but
# - at some point we'll want to get rid of all the python2 compatibility code
# - it would be duplicating the standard library's approach which has surely had more people looking at it
# i have tried to use close-stdout approach and fix all stale references, however new ones keep popping
# up e.g. in logging, unittest, curses. there are probably more, which means hard-to-debug future crashes
# the only reason to redirect c-level stream were:
# - pyaudio on alsa produces a lot of chatter on init which is annoying on a console interface
# - pygame prints a "banner" message to the console when imported, which is annoying especially when using another interface
# since:
# - the first is annoying but does not deface the screen during runtime
# - the second only occurs in a deprecated interface, and can be avoided in other interfaces by delaying module load until it's needed
# - the whole thing is a bit hacky and clearly likely to break in newer python versions
# - there are probably still crashes hidden away in some code paths
# all of this is an unreasonable maintenance load and it is better to go with the supported method
# of doing a "shallow" redirect by reassigning sys.std***.


class StdIOBase(object):
    """Holds standard unicode streams."""

    def __init__(self):
        self._attach_stdin()
        self._attach_output_stream('stdout')
        self._attach_output_stream('stderr')

    def _attach_stdin(self):
        # stdio becomes None in GUI mode
        # use __stdin__ as we depend elsewhere on this having a .buffer and pointing to true stdin
        self.stdin = sys.__stdin__ or _open_named_devnull('<stdin>', 'r')

    def _attach_output_stream(self, stream_name, redirected=False):
        # stdio becomes None in GUI mode
        stream = getattr(sys, '__%s__' % (stream_name,))
        if not stream:
            stream = _open_named_devnull('<%s>' % (stream_name,), 'w')
        setattr(self, stream_name, stream)

    # unicode stream wrappers

    def _wrap_output_stream(self, stream, encoding=None):
        """Wrap std bytes streams or redirected files to make them behave more like in Python 3."""
        encoding = encoding or stream.encoding or 'utf-8'
        wrapped = codecs.getwriter(encoding)(stream, errors='replace')
        wrapped.buffer = stream
        return wrapped

    def _wrap_input_stream(self, stream, encoding=None):
        """Wrap std bytes streams or redirected files to make them behave more like in Python 3."""
        encoding = encoding or stream.encoding or 'utf-8'
        wrapped = codecs.getreader(encoding)(stream)
        wrapped.buffer = stream
        return wrapped

    @contextmanager
    def _muffle_one(self, stream_name, preserve, redirect):
        """Silence stdout or stderr. On Python 2, also silences external writes."""
        std_stream = getattr(sys, '__%s__' % (stream_name,))
        if not PY2:
            encoding = std_stream.encoding
        save = None
        try:
            if not PY2:
                save_buffer = std_stream.buffer
            else: # pragma: no cover
                try:
                    # save the file descriptor for the target stream
                    save = os.dup(std_stream.fileno())
                except EnvironmentError as e:
                    # won't work, give up
                    logging.error(e)
                    yield
                    return
            if preserve or redirect:
                temp_file = tempfile.TemporaryFile('w+b')
            else:
                temp_file = io.open(os.devnull, 'wb')
            with temp_file as temp:
                std_stream.flush()
                if not PY2:
                    # replace the underlying buffer in the TextIOWrapper
                    # since we're mutating the object, affects all stored references.
                    std_stream.__init__(temp_file, encoding=encoding) # errors='replace'
                else: # pragma: no cover
                    # the old unix way, doesn't work on python3.6+/windows
                    # put /dev/null fds on 1 (stdout) or 2 (stderr)
                    os.dup2(temp.fileno(), std_stream.fileno())
                # fix our own streams - needed on Windows
                # as we keep a hacked stream writing to the Windows API
                self._attach_output_stream(stream_name, redirected=True)
                # do stuff
                try:
                    yield
                finally:
                    std_stream.flush()
                    # restore file descriptors
                    if not PY2:
                        std_stream.__init__(save_buffer, encoding=encoding)
                    else: # pragma: no cover
                        os.dup2(save, std_stream.fileno())
                    self._attach_output_stream(stream_name, redirected=False)
                    if preserve:
                        if PY2: # pragma: no cover
                            redirect = std_stream
                        else:
                            redirect = std_stream.buffer
                    if redirect:
                        # write contents of temporary file into redirect or back into standard io
                        temp.flush()
                        temp.seek(0)
                        redirect.write(temp.read())
        finally:
            if save is not None:
                os.close(save)

    @contextmanager
    def _muffle(self, stream_name, **kwargs):
        """Silence stdout, stderr, or both."""
        if not stream_name:
            with self._muffle_one('stdout', **kwargs):
                with self._muffle_one('stderr', **kwargs):
                    yield
        else:
            with self._muffle_one(stream_name, **kwargs):
                yield

    @contextmanager
    def pause(self, stream_name=None):
        """Pause stdout or stderr or both, preserving output."""
        with self._muffle(stream_name, preserve=True, redirect=None):
            yield

    @contextmanager
    def quiet(self, stream_name=None):
        """Pause stdout or stderr or both, not preserving output."""
        with self._muffle(stream_name, preserve=False, redirect=None):
            yield

    @contextmanager
    def redirect_output(self, output=None, stream_name=None):
        """Pause stdout or stderr or both, redirecting output."""
        with self._muffle(stream_name, preserve=False, redirect=output):
            yield



"""
PC-BASIC - files.py
Devices, Files and I/O operations

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import os
import sys
import logging
import io

from ...compat import xrange, int2byte, text_type
from ...compat import iterchar, iteritems, getcwdu
from ...compat import split_quoted

from ..base import error
from ..base import tokens as tk
from .. import values
from . import formatter
from . import devicebase
from . import cassette
from . import disk
from . import ports
from . import parports


# MS-DOS device files
DOS_DEVICE_FILES = (b'AUX', b'CON', b'NUL', b'PRN')

# allowable drive letters in GW-BASIC are letters or @
DRIVE_LETTERS = b'@' + tk.UPPERCASE


############################################################################
# General file manipulation

class Files(object):
    """File manager."""

    def __init__(
            self, values, memory, queues, keyboard, display, console,
            max_files, max_reclen, serial_buffer_size,
            device_params, current_device,
            codepage, text_mode, soft_linefeed
        ):
        """Initialise files."""
        # for wait() in files_
        self._queues = queues
        self._values = values
        self._memory = memory
        self.files = {}
        self.max_files = max_files
        self.max_reclen = max_reclen
        self._init_devices(
            values, queues, display, console, keyboard,
            device_params, current_device,
            serial_buffer_size, codepage, text_mode, soft_linefeed
        )

    ###########################################################################
    # file management

    def close(self, num):
        """Close a numbered file."""
        try:
            self.files[num].close()
            del self.files[num]
        except KeyError:
            pass

    def close_all(self):
        """Close all files."""
        for f in self.files.values():
            f.close()
        self.files = {}

    def open(
            self, number, description, filetype, mode=b'I', access=b'', lock=b'',
            reclen=128, seg=0, offset=0, length=0
        ):
        """Open a file on a device specified by description."""
        if (not description) or (number < 0) or (number > self.max_files):
            # bad file number; also for name='', for some reason
            raise error.BASICError(error.BAD_FILE_NUMBER)
        if number in self.files:
            raise error.BASICError(error.FILE_ALREADY_OPEN)
        mode = mode.upper()
        device, dev_param = self._get_device_param(description, mode)
        # get the field buffer
        field = self._memory.fields[number] if number else None
        # open the file on the device
        new_file = device.open(
            number, dev_param, filetype, mode, access, lock,
            reclen, seg, offset, length, field
        )
        logging.debug(
            'Opened file %r as #%d (type %s, mode %s)', dev_param, number, filetype, mode
        )
        if number:
            self.files[number] = new_file
        return new_file

    def get(self, num, mode=b'IOAR', not_open=error.BAD_FILE_NUMBER):
        """Get the file object for a file number and check allowed mode."""
        if (num < 1):
            raise error.BASICError(error.BAD_FILE_NUMBER)
        try:
            the_file = self.files[num]
        except KeyError:
            raise error.BASICError(not_open)
        if the_file.mode.upper() not in mode:
            raise error.BASICError(error.BAD_FILE_MODE)
        return the_file

    def _get_from_integer(self, num, mode=b'IOAR'):
        """Get the file object for an Integer file number and check allowed mode."""
        num = values.to_int(num, unsigned=True)
        error.range_check(0, 255, num)
        return self.get(num, mode)

    ###########################################################################
    # device management

    def _init_devices(
            self, values, queues, display, console, keyboard,
            device_params, current_device,
            serial_in_size, codepage, text_mode, soft_linefeed
        ):
        """Initialise devices."""
        device_params = self._normalise_params(device_params)
        # screen device, for files_()
        current_device = self._normalise_current_device(current_device, device_params)
        self._console = console
        self._keyboard = keyboard
        self._devices = {
            b'SCRN:': devicebase.SCRNDevice(display, console),
            # KYBD: device needs display as it can set the screen width
            b'KYBD:': devicebase.KYBDDevice(keyboard, display),
            # cassette: needs text screen to display Found and Skipped messages
            b'CAS1:': cassette.CASDevice(device_params.get(b'CAS1', None), self._console),
            # serial devices
            b'COM1:': ports.COMDevice(device_params.get(b'COM1', None), queues, serial_in_size),
            b'COM2:': ports.COMDevice(device_params.get(b'COM2', None), queues, serial_in_size),
            # parallel devices - LPT1: must always be available
            b'LPT1:': parports.LPTDevice(
                device_params.get(b'LPT1', None), devicebase.nullstream(), codepage
            ),
            b'LPT2:': parports.LPTDevice(device_params.get(b'LPT2', None), None, codepage),
            b'LPT3:': parports.LPTDevice(device_params.get(b'LPT3', None), None, codepage),
        }
        # device files
        self.scrn_file = self._devices[b'SCRN:'].device_file
        self.kybd_file = self._devices[b'KYBD:'].device_file
        self.lpt1_file = self._devices[b'LPT1:'].device_file
        # disks
        self._init_disk_devices(device_params, current_device, codepage, text_mode, soft_linefeed)

    def _normalise_current_device(self, current_device, device_params):
        """Normalise current device specification."""
        current_device = self._normalise_device_name(current_device)
        if current_device and current_device != b'CAS1' and current_device not in DRIVE_LETTERS:
            logging.error('Invalid current device `%s`', current_device)
            current_device = b''
        # no current device specified, or set to an undefined device
        if (
                not current_device
                or current_device not in device_params.keys()
                or not device_params[current_device]
            ):
            if b'Z' in device_params and device_params[b'Z']:
                # use z: if this is defined
                current_device = b'Z'
            else:
                # otherwise, set current device to last disk available
                # if nothing available, use the internal drive @
                available = sorted(
                    _k for _k in device_params.keys() if len(_k) == 1 and device_params[_k]
                )
                if available:
                    current_device = available[-1]
                else:
                    current_device = b'@'
        return current_device

    def _normalise_params(self, device_params):
        """Normalise keys in device parameter dict."""
        if not device_params:
            return {}
        output_dict = {}
        for key, value in iteritems(device_params):
            key = self._normalise_device_name(key)
            if not key:
                continue
            # convert value to unicode
            if isinstance(value, bytes):
                try:
                    value = value.decode('ascii')
                except UnicodeError:
                    logging.error(
                        'Invalid device parameter value: `%s` must be ascii if given as bytes.', key
                    )
                    continue
            output_dict[key] = value
        return output_dict

    def _normalise_device_name(self, key):
        """Normalise device name to uppercase ascii bytes without colon."""
        if not key:
            return b''
        if isinstance(key, text_type):
            try:
                key = key.encode('ascii')
            except UnicodeError:
                logging.error('Invalid device name: `%s` is not ascii.', key)
                return b''
        # strip off trailing : if provided
        if key.endswith(b':'):
            key = key[:-1]
        # convert to uppercase
        key = key.upper()
        return key

    def close_devices(self):
        """Close device master files."""
        for d in self._devices.values():
            d.close()

    def device_available(self, spec):
        """Return whether the device indicated by the spec (including :) is available."""
        dev_name = spec.split(b':', 1)[0] + b':'
        return (dev_name in self._devices) and self._devices[dev_name].available()

    def get_device(self, name):
        """Get a device by name (including :) or KeyError if not there."""
        return self._devices[name]

    def _get_device_param(self, file_spec, mode):
        """Get a device object and parameters from a file specification."""
        name = bytes(file_spec)
        split = name.split(b':', 1)
        if len(split) > 1:
            # colon (:) found
            dev_name = split[0].upper() + b':'
            dev_param = split[1]
            try:
                device = self._devices[dev_name]
            except KeyError:
                # not an allowable device or drive name
                # bad file number, for some reason
                raise error.BASICError(error.BAD_FILE_NUMBER)
        else:
            device = self._devices[self._current_device + b':']
            # MS-DOS device aliases - these can't be names of disk files
            if device != self._devices[b'CAS1:'] and name in DOS_DEVICE_FILES:
                if name == b'AUX':
                    device, dev_param = self._devices[b'COM1:'], b''
                elif name == b'CON' and mode == b'I':
                    device, dev_param = self._devices[b'KYBD:'], b''
                elif name == b'CON' and mode == b'O':
                    device, dev_param = self._devices[b'SCRN:'], b''
                elif name == b'PRN':
                    device, dev_param = self._devices[b'LPT1:'], b''
                elif name == b'NUL':
                    device, dev_param = devicebase.NullDevice(), b''
            else:
                # open file on default device
                dev_param = name
        return device, dev_param

    ###########################################################################
    # statement callbacks

    def reset_(self, args):
        """RESET: Close all files."""
        list(args)
        self.close_all()

    def close_(self, args):
        """CLOSE: close a file, or all files."""
        at_least_one = False
        for number in args:
            number = values.to_int(number)
            error.range_check(0, 255, number)
            at_least_one = True
            # close() deals with non-open numbers
            self.close(number)
        # if no file number given, close everything
        if not at_least_one:
            self.close_all()

    def open_(self, args):
        """OPEN: open a data file."""
        first_expr = values.next_string(args)
        if next(args):
            # old syntax
            mode = first_expr[:1].upper()
            if mode not in (b'I', b'O', b'A', b'R'):
                raise error.BASICError(error.BAD_FILE_MODE)
            number = values.to_int(next(args))
            error.range_check(0, 255, number)
            name = values.next_string(args)
            access, lock = None, None
        else:
            # new syntax
            name = first_expr
            mode, access, lock = next(args), next(args), next(args)
            # AS file number clause
            number = values.to_int(next(args))
            error.range_check(0, 255, number)
        reclen, = args
        mode = mode or b'R'
        if reclen is None:
            reclen = 128
        else:
            reclen = values.to_int(reclen)
        # mode and access (if specified) must match if not a RANDOM file
        # If FOR APPEND ACCESS WRITE is specified, raises PATH/FILE ACCESS ERROR
        # If FOR and ACCESS mismatch in other ways, raises SYNTAX ERROR.
        if access:
            if mode == b'A' and access == b'W':
                raise error.BASICError(error.PATH_FILE_ACCESS_ERROR)
            elif ((mode == b'I' and access != b'R') or (mode == b'O' and access != b'W') or
                    (mode == b'A' and access != b'RW')):
                raise error.BASICError(error.STX)
        error.range_check(1, self.max_reclen, reclen)
        # can't open file 0, or beyond max_files
        error.range_check_err(1, self.max_files, number, error.BAD_FILE_NUMBER)
        self.open(number, name, b'D', mode=mode, access=access, lock=lock, reclen=reclen)

    ###########################################################################

    def field_(self, args):
        """FIELD: attach a variable to the record buffer."""
        number = values.to_int(next(args))
        error.range_check(0, 255, number)
        # check if file is open
        self.get(number, b'R')
        offset = 0
        try:
            while True:
                width = values.to_int(next(args))
                error.range_check(0, 255, width)
                name, index = next(args)
                name = self._memory.complete_name(name)
                self._memory.fields[number].attach_var(name, index, offset, width)
                offset += width
        except StopIteration:
            pass

    def _check_pos(self, pos):
        """Helper function: PUT and GET syntax."""
        if pos is None:
            return pos
        # forcing to single before rounding - this means we don't have enough precision
        # to address each individual record close to the maximum record number
        # but that's in line with GW
        pos = int(round(values.to_single(pos).to_value()))
        # not 2^32-1 as the manual boasts!
        # pos-1 needs to fit in a single-precision mantissa
        error.range_check_err(1, 2**25, pos, err=error.BAD_RECORD_NUMBER)
        return pos

    def put_(self, args):
        """PUT: write record to file."""
        number = values.to_int(next(args))
        error.range_check(0, 255, number)
        the_file = self.get(number, b'R', not_open=error.BAD_FILE_MODE)
        pos, = args
        pos = self._check_pos(pos)
        the_file.put(pos)

    def get_(self, args):
        """GET: read record from file."""
        number = values.to_int(next(args))
        error.range_check(0, 255, number)
        the_file = self.get(number, b'R', not_open=error.BAD_FILE_MODE)
        pos, = args
        pos = self._check_pos(pos)
        the_file.get(pos)

    ###########################################################################

    def _get_lock_limits(self, lock_start_rec, lock_stop_rec):
        """Get record lock limits."""
        if lock_start_rec is None and lock_stop_rec is None:
            return None, None
        if lock_start_rec is None:
            lock_start_rec = 1
        else:
            lock_start_rec = round(values.to_single(lock_start_rec).to_value())
        if lock_stop_rec is None:
            lock_stop_rec = lock_start_rec
        else:
            lock_stop_rec = round(values.to_single(lock_stop_rec).to_value())
        if lock_start_rec < 1 or lock_start_rec > 2**25-2 or lock_stop_rec < 1 or lock_stop_rec > 2**25-2:
            raise error.BASICError(error.BAD_RECORD_NUMBER)
        return lock_start_rec, lock_stop_rec

    def lock_(self, args):
        """LOCK: set file or record locks."""
        num = values.to_int(next(args))
        error.range_check(0, 255, num)
        thefile = self.get(num)
        lock_start_rec, lock_stop_rec = args
        try:
            thefile.lock(*self._get_lock_limits(lock_start_rec, lock_stop_rec))
        except AttributeError:
            # not a disk file
            raise error.BASICError(error.PERMISSION_DENIED)

    def unlock_(self, args):
        """UNLOCK: set file or record locks."""
        num = values.to_int(next(args))
        error.range_check(0, 255, num)
        thefile = self.get(num)
        lock_start_rec, lock_stop_rec = args
        try:
            thefile.unlock(*self._get_lock_limits(lock_start_rec, lock_stop_rec))
        except AttributeError:
            # not a disk file
            raise error.BASICError(error.PERMISSION_DENIED)

    ###########################################################################

    def write_(self, args):
        """WRITE: Output machine-readable expressions to the screen or a file."""
        file_number = next(args)
        if file_number is None:
            output = self.scrn_file
        else:
            file_number = values.to_int(file_number)
            error.range_check(0, 255, file_number)
            output = self.get(file_number, b'OAR')
        outstrs = []
        try:
            while True:
                expr = next(args)
                if isinstance(expr, values.String):
                    outstrs.append(b'"%s"' % expr.to_str())
                else:
                    outstrs.append(values.to_repr(expr, leading_space=False, type_sign=False))
        except StopIteration:
            # write the whole thing as one thing (this affects line breaks)
            output.write_line(b','.join(outstrs))
        except error.BASICError:
            if outstrs:
                output.write(b','.join(outstrs) + b',')
            raise

    def width_(self, args):
        """WIDTH: set width of screen or device."""
        file_or_device = next(args)
        num_rows_dummy = None
        if file_or_device == tk.LPRINT:
            dev = self.lpt1_file
            w = values.to_int(next(args))
        elif isinstance(file_or_device, values.Number):
            file_or_device = values.to_int(file_or_device)
            error.range_check(0, 255, file_or_device)
            dev = self.get(file_or_device, mode=b'IOAR')
            w = values.to_int(next(args))
        else:
            expr = next(args)
            if isinstance(expr, values.String):
                devname = expr.to_str().upper()
                w = values.to_int(next(args))
                try:
                    dev = self._devices[devname].device_file
                except (KeyError, AttributeError):
                    # bad file name
                    raise error.BASICError(error.BAD_FILE_NAME)
            else:
                w = values.to_int(expr)
                num_rows_dummy = next(args)
                if num_rows_dummy is not None:
                    num_rows_dummy = values.to_int(num_rows_dummy)
                dev = self.scrn_file
        error.range_check(0, 255, w)
        list(args)
        if num_rows_dummy is not None:
            self.scrn_file._display.set_height(num_rows_dummy)
        dev.set_width(w)

    def print_(self, args):
        """PRINT: Write expressions to the screen or a file."""
        # check for a file number
        file_number = next(args)
        if file_number is not None:
            file_number = values.to_int(file_number)
            error.range_check(0, 255, file_number)
            output = self.get(file_number, b'OAR')
            console = None
        else:
            # neither LPRINT not a file number: print to screen
            output = self.scrn_file
            console = self.scrn_file.console
        formatter.Formatter(output, console).format(args)

    def lprint_(self, args):
        """LPRINT: Write expressions to printer LPT1."""
        formatter.Formatter(self.lpt1_file).format(args)

    ###########################################################################

    def ioctl_statement_(self, args):
        """IOCTL: send control string to I/O device. Not implemented."""
        num = values.to_int(next(args))
        error.range_check(0, 255, num)
        thefile = self.get(num)
        control_string = values.next_string(args)
        list(args)
        logging.warning('IOCTL statement not implemented.')
        raise error.BASICError(error.IFC)

    def motor_(self, args):
        """MOTOR: drive cassette motor; not implemented."""
        logging.warning('MOTOR statement not implemented.')
        val = next(args)
        if val is not None:
            error.range_check(0, 255, values.to_int(val))
        list(args)

    def lcopy_(self, args):
        """LCOPY: screen copy / no-op in later GW-BASIC."""
        # See e.g. http://shadowsshot.ho.ua/docs001.htm#LCOPY
        val = next(args)
        if val is not None:
            error.range_check(0, 255, values.to_int(val))
        list(args)

    ###########################################################################
    # function callbacks

    def loc_(self, args):
        """LOC: get file pointer."""
        num, = args
        num = values.to_integer(num)
        loc = self._get_from_integer(num).loc()
        return self._values.new_single().from_int(loc)

    def eof_(self, args):
        """EOF: get end-of-file."""
        num, = args
        num = values.to_integer(num)
        eof = self._values.new_integer()
        if not num.is_zero() and self._get_from_integer(num, b'IR').eof():
            eof = eof.from_int(-1)
        return eof

    def lof_(self, args):
        """LOF: get length of file."""
        num, = args
        num = values.to_integer(num)
        lof = self._get_from_integer(num).lof()
        return self._values.new_single().from_int(lof)

    def lpos_(self, args):
        """LPOS: get the current printer column."""
        num, = args
        num = values.to_int(num)
        error.range_check(0, 3, num)
        printer = self._devices[b'LPT%d:' % max(1, num)]
        col = printer.device_settings.col
        # follow weird GW-BASIC behaviour
        # this is reported as 1 if it equals the DEVICE's width plus one
        # even if it then continues until the FILE's width afterwards
        if printer.device_file and col == printer.device_file.width + 1:
            col = 1
        return self._values.new_integer().from_int(col % 256)

    def input_(self, args):
        """INPUT$: read num chars from file or keyboard."""
        num = values.to_int(next(args))
        error.range_check(1, 255, num)
        filenum = next(args)
        if filenum is not None:
            filenum = values.to_int(filenum)
            error.range_check(0, 255, filenum)
            # raise BAD FILE MODE (not BAD FILE NUMBER) if the file is not open
            read = self.get(filenum, mode=b'IR', not_open=error.BAD_FILE_MODE).read
        else:
            read = self._keyboard.read_bytes_block
        list(args)
        # read the chars
        word = read(num)
        if len(word) < num:
            # input past end
            raise error.BASICError(error.INPUT_PAST_END)
        return self._values.new_string().from_str(word)


    ###########################################################################

    def ioctl_(self, args):
        """IOCTL$: read device control string response; not implemented."""
        num = values.to_int(next(args))
        error.range_check(0, 255, num)
        # raise BAD FILE NUMBER if the file is not open
        infile = self.get(num)
        list(args)
        logging.warning('IOCTL$ function not implemented.')
        raise error.BASICError(error.IFC)

    def erdev_(self, args):
        """ERDEV: device error value; not implemented."""
        list(args)
        logging.warning('ERDEV function not implemented.')
        return self._values.new_integer()

    def erdev_str_(self, args):
        """ERDEV$: device error string; not implemented."""
        list(args)
        logging.warning('ERDEV$ function not implemented.')
        return self._values.new_string()

    def exterr_(self, args):
        """EXTERR: device error information; not implemented."""
        val, = args
        logging.warning('EXTERR function not implemented.')
        error.range_check(0, 3, values.to_int(val))
        return self._values.new_integer()


    ###########################################################################
    # disk devices

    def _init_disk_devices(
            self, device_params, current_device,
            codepage, text_mode, soft_linefeed
        ):
        """Initialise disk devices."""
        # if Z not specified, mount to cwd by default (override by specifying 'Z': None)
        if b'Z' not in device_params:
            device_params[b'Z'] = getcwdu()
        # disk devices
        for letter in iterchar(DRIVE_LETTERS):
            if letter in device_params and device_params[letter]:
                # drive can be non-empty only on Windows, needs to be split out first as we use :
                drive, drivepath = os.path.splitdrive(device_params[letter])
                params = split_quoted(
                    drivepath, split_by=u':', quote=u'"', strip_quotes=True
                )
                path = drive + params[0]
                if len(params) > 1:
                    cwd = params[1]
                    # ignore any further specifiers
                else:
                    cwd = u''
            else:
                path, cwd = u'', u''
            # treat device @: separately - internal disk must exist but may remain unmounted
            disk_class = disk.InternalDiskDevice if letter == b'@' else disk.DiskDevice
            self._devices[letter + b':'] = disk_class(
                letter, path, cwd, codepage, text_mode, soft_linefeed
            )
        # current_device value is normalised
        self._current_device = current_device

    def _get_diskdevice_and_path(self, path):
        """Return the disk device and remaining path for given file spec."""
        # careful - do not convert path to uppercase, we still need to match
        splits = bytes(path).split(b':', 1)
        if len(splits) == 0:
            dev, spec = self._current_device, b''
        elif len(splits) == 1:
            dev, spec = self._current_device, splits[0]
        else:
            try:
                dev, spec = splits[0].upper(), splits[1]
            except KeyError:
                raise error.BASICError(error.DEVICE_UNAVAILABLE)
        # must be a disk device
        if dev not in DRIVE_LETTERS:
            raise error.BASICError(error.DEVICE_UNAVAILABLE)
        return self._devices[dev + b':'], spec

    def get_native_cwd(self):
        """Get current working directory on current drive."""
        # must be a disk device
        if self._current_device not in DRIVE_LETTERS:
            raise error.BASICError(error.IFC)
        return self._devices[self._current_device + b':'].get_native_cwd()

    def chdir_(self, args):
        """CHDIR: change working directory."""
        name = values.next_string(args)
        list(args)
        if not name:
            raise error.BASICError(error.BAD_FILE_NAME)
        dev, path = self._get_diskdevice_and_path(name)
        dev.chdir(path)

    def mkdir_(self, args):
        """MKDIR: create directory."""
        name = values.next_string(args)
        list(args)
        if not name:
            raise error.BASICError(error.BAD_FILE_NAME)
        dev, path = self._get_diskdevice_and_path(name)
        dev.mkdir(path)

    def rmdir_(self, args):
        """RMDIR: remove directory."""
        name = values.next_string(args)
        list(args)
        if not name:
            raise error.BASICError(error.BAD_FILE_NAME)
        dev, path = self._get_diskdevice_and_path(name)
        dev.rmdir(path)

    def name_(self, args):
        """NAME: rename file or directory."""
        dev, oldpath = self._get_diskdevice_and_path(values.next_string(args))
        # don't rename open files
        # NOTE: we need to check file exists before parsing the next name
        # to get the same error sequencing as GW-BASIC
        dev.require_file_exists(oldpath)
        dev.require_file_not_open(oldpath)
        newdev, newpath = self._get_diskdevice_and_path(values.next_string(args))
        dev.require_file_not_open(newpath)
        list(args)
        if dev != newdev:
            raise error.BASICError(error.RENAME_ACROSS_DISKS)
        dev.rename(oldpath, newpath)

    def kill_(self, args):
        """KILL: remove file."""
        name = values.next_string(args)
        list(args)
        if not name:
            raise error.BASICError(error.BAD_FILE_NAME)
        dev, path = self._get_diskdevice_and_path(name)
        dev.kill(path)

    def files_(self, args):
        """FILES: output directory listing to screen."""
        pathmask = values.next_string(args)
        list(args)
        # pathmask may be left unspecified, but not empty
        if pathmask == b'':
            raise error.BASICError(error.BAD_FILE_NAME)
        elif pathmask is None:
            pathmask = b''
        dev, path = self._get_diskdevice_and_path(pathmask)
        # retrieve files first (to ensure correct path/file not found errors)
        output = dev.listdir(path)
        num_cols = self._console.width // 20
        # output working dir in DOS format
        # NOTE: this is always the current dir, not the one being listed
        self._console.write_line(dev.get_cwd())
        if not output:
            raise error.BASICError(error.FILE_NOT_FOUND)
        # output files
        for i, cols in enumerate(output[j:j+num_cols] for j in xrange(0, len(output), num_cols)):
            self._console.write_line(b' '.join(cols))
            if not (i % 4):
                # allow to break during dir listing & show names flowing on screen
                self._queues.wait()
            i += 1
        self._console.write_line(b' %d Bytes free\n' % dev.get_free())



"""
PC-BASIC - compat.base
Cross-platform compatibility utilities

(c) 2018--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import os
import sys
import platform


# Python major version
PY2 = sys.version_info.major == 2

# platform constants
WIN32 = sys.platform == 'win32'
MACOS = sys.platform == 'darwin'

# 64-bit (needed for Windows binary modules)
X64 = platform.architecture()[0] == '64bit'

# platform tag for libraries
if WIN32:
    PLATFORM = sys.platform + ('_x64' if X64 else '_x86')
else:
    PLATFORM = sys.platform

# user configuration and state directories
HOME_DIR = os.path.expanduser(u'~')

if WIN32:
    USER_CONFIG_HOME = os.getenv(u'APPDATA', default=u'')
    USER_DATA_HOME = USER_CONFIG_HOME
elif MACOS:
    USER_CONFIG_HOME = os.path.join(HOME_DIR, u'Library', u'Application Support')
    USER_DATA_HOME = USER_CONFIG_HOME
else:
    USER_CONFIG_HOME = os.environ.get(u'XDG_CONFIG_HOME') or os.path.join(HOME_DIR, u'.config')
    USER_DATA_HOME = os.environ.get(u'XDG_DATA_HOME') or os.path.join(HOME_DIR, u'.local', u'share')

# package/executable directory
if hasattr(sys, 'frozen'):
    # we're a package: get the directory of the packaged executable
    # (__file__ is undefined in frozen packages)
    # this is for cx_Freeze's package layout
    BASE_DIR = os.path.join(os.path.dirname(sys.executable), 'lib', 'pcbasic')
else:
    BASE_DIR = os.path.abspath(os.path.join(os.path.dirname(os.path.realpath(__file__)), '..'))



"""
PC-BASIC - bytematrix.py
2D matrices of bytes

(c) 2018--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import operator
from binascii import hexlify, unhexlify

from ...compat import zip, int2byte, xrange, iterbytes, iterchar


class ByteMatrix(object):
    """2D byte matrix."""

    def __init__(self, height=0, width=0, data=0):
        """Create a new matrix."""
        self._height = height
        self._width = width
        if not width and not height:
            self._rows = [bytearray()]
        elif isinstance(data, int):
            self._rows = [bytearray([data])*width for _ in xrange(self._height)]
        else:
            # assume iterable, TypeError if not
            data = list(data)
            if len(data) == height:
                if isinstance(data[0], int):
                    # bytearrays and python3 bytes
                    self._rows = [bytearray([_row]) for _row in data]
                else:
                    assert len(data[0]) == width
                    self._rows = [bytearray(_row) for _row in data]
            else:
                assert len(data) == height * width
                self._rows = [
                    bytearray(data[_offs : _offs+width])
                    for _offs in xrange(0, len(data), width)
                ]

    def __repr__(self):
        """Debugging representation."""
        hexreps = [''.join('\\x{:02x}'.format(_c) for _c in _row) for _row in self._rows]
        return "ByteMatrix({0._height}, {0._width}, [\n    '{1}' ])".format(
            self, "',\n    '".join(hexreps)
        )

    def __getitem__(self, index):
        """Extract items by [y, x] indexing or slicing or 1D index."""
        y, x = index
        if isinstance(y, slice):
            if isinstance(x, slice):
                return self._create_from_rows([_row[x] for _row in self._rows[y]])
            else:
                return self._create_from_rows([bytearray([_row[x]]) for _row in self._rows[y]])
        if isinstance(x, slice):
            return self._create_from_rows([self._rows[y][x]])
        return self._rows[y][x]

    def __setitem__(self, index, value):
        """Set items by [y, x] indexing or slicing."""
        y, x = index
        if isinstance(value, int):
            if isinstance(x, slice):
                if isinstance(y, slice):
                    for row in self._rows[y]:
                        row[x] = bytearray(value for _ in iterbytes(row[x]))
                else:
                    self._rows[y][x] = bytearray(value for _ in iterbytes(self._rows[y][x]))
            else:
                if isinstance(y, slice):
                    for row in self._rows[y]:
                        row[x] = value
                else:
                    self._rows[y][x] = value
        else:
            if isinstance(value, ByteMatrix):
                value = value._rows
            elif type(value) not in (list, int):
                raise TypeError(
                    'Can only assign ByteMatrix, list of bytes-like or int, not %s.' % type(value)
                )
            if isinstance(y, slice):
                if isinstance(x, slice):
                    # this will fail if we're self-assigning a slice of a view to the original view
                    # as we'll be overwriting the source while writing to the destination.
                    # in those cases, we'll need to copy first
                    for _dst, _src in zip(self._rows[y], value):
                        # if x is a slice, this will copy too -> e.g. array[:] = another_array
                        # but not if rhs is a view?
                        _dst[x] = bytearray(_src)
                else:
                    for _dst, _src in zip(self._rows[y], value):
                        # if x is a slice, this will copy too -> e.g. array[:] = another_array
                        # but not if rhs is a view?
                        _dst[x] = _src[0]
            elif isinstance(x, slice):
                assert len(value) == 1
                self._rows[y][x] = value[0]

    def __eq__(self, rhs):
        """Equality to other byte matrix."""
        # do quick checks first
        return self.width == rhs.width and self.height == rhs.height and self._rows == rhs._rows

    def __ne__(self, rhs):
        """Non-equality to other byte matrix."""
        return not self.__eq__(rhs)

    def _elementwise_list(self, rhs, oper):
        """Helper for elementwise operations."""
        if isinstance(rhs, int):
            return [
                bytearray(oper(_lbyte, rhs) for _lbyte in iterbytes(_lrow))
                for _lrow in self._rows
            ]
        else:
            assert self._height == rhs._height
            assert self._width == rhs._width
            return [
                bytearray(
                    oper(_lbyte, _rbyte)
                    for _lbyte, _rbyte in zip(iterbytes(_lrow), iterbytes(_rrow))
                )
                for _lrow, _rrow in zip(self._rows, rhs._rows)
            ]

    def elementwise(self, rhs, oper):
        """Element-wise operation with another matrix or a scalar."""
        return self._create_from_rows(self._elementwise_list(rhs, oper))

    def __or__(self, rhs):
        """Bitwise or."""
        return self.elementwise(rhs, operator.__or__)

    def __and__(self, rhs):
        """Bitwise and."""
        return self.elementwise(rhs, operator.__and__)

    def __xor__(self, rhs):
        """Bitwise exclusive or."""
        return self.elementwise(rhs, operator.__xor__)

    def __rshift__(self, rhs):
        """Right-shift."""
        return self.elementwise(rhs, operator.__rshift__)

    def __lshift__(self, rhs):
        """Byte-masked left-shift."""
        return self.elementwise(rhs, lambda _l, _r: (_l << _r) & 0xff)

    def elementwise_inplace(self, rhs, oper):
        """In-place element-wise operation with another matrix or a scalar."""
        self._rows = self._elementwise_list(rhs, oper)
        return self

    def __ior__(self, rhs):
        """In-place bitwise or."""
        return self.elementwise_inplace(rhs, operator.__ior__)

    def __iand__(self, rhs):
        """In-place bitwise and."""
        return self.elementwise_inplace(rhs, operator.__iand__)

    def __ixor__(self, rhs):
        """In-place bitwise exclusive or."""
        return self.elementwise_inplace(rhs, operator.__ixor__)

    def __irshift__(self, rhs):
        """In-place right-shift."""
        return self.elementwise_inplace(rhs, operator.__irshift__)

    def __ilshift__(self, rhs):
        """In-place left-shift."""
        return self.elementwise_inplace(rhs, lambda _l, _r: (_l << _r) & 0xff)

    @property
    def width(self):
        """Number of columns."""
        return self._width

    @property
    def height(self):
        """Number of rows."""
        return self._height

    @classmethod
    def _create_from_rows(cls, data):
        """Construct byte matrix from rows of bytearrays."""
        new = cls()
        if not data:
            new._height = len(data)
            new._width = 0
            new._rows = [bytearray()]
        else:
            new._height = len(data)
            new._width = len(data[0])
            new._rows = data
        assert len(set(len(_r) for _r in new._rows)) == 1, 'ByteMatrix rows must all be same length'
        return new

    @classmethod
    def frompacked(cls, packed, height, items_per_byte):
        """Unpack from packed-bits representation."""
        packed = bytes(bytearray(packed))
        if not packed or not height:
            return cls(0, 0)
        width = len(packed) // height
        if not width:
            return cls(0, 0)
        return cls._create_from_rows([
            unpack_bytes(packed[_offs : _offs+width], items_per_byte)
            for _offs in xrange(0, len(packed), width)
        ])

    def packed(self, items_per_byte):
        """Pack into packed-bits representation, byte aligned on rows."""
        return bytearray().join(
            pack_bytes(_r, items_per_byte) for _r in self._rows
        )

    @classmethod
    def fromhex(cls, hex, height, items_per_byte):
        """Unpack from hex representation."""
        return cls.frompacked(unhexlify(hex), height, items_per_byte)

    def hex(self, items_per_byte):
        """Pack to hex representation."""
        return hexlify(self.packed(items_per_byte))

    def render(self, back, fore):
        """Set attributes on bit matrix."""
        return self._create_from_rows([
            bytearray(fore if _c else back for _c in _row)
            for _row in self._rows
        ])

    def hextend(self, by_width, fill=0):
        """Extend width by given number of bytes."""
        new_row = bytearray([fill])*by_width
        return self._create_from_rows([_row + new_row for _row in self._rows])

    def vextend(self, by_height, fill=0):
        """Extend height by given number of bytes."""
        return self._create_from_rows(
            self._rows
            + [bytearray([fill])*self._width for _ in range(by_height)]
        )

    def hrepeat(self, times=1):
        """Multiply width by byte repetition (00 11 22 ...)."""
        return self._create_from_rows([
            bytearray(_byte for _byte in _row for _ in range(times))
            for _row in self._rows
        ])

    def vrepeat(self, times=1):
        """Multiply height by row repetition."""
        return self._create_from_rows([
            bytearray(_row) for _row in self._rows
            for _ in range(times)
        ])

    def htile(self, times=1):
        """Multiply width by tiling (012 012 ...)."""
        return self._create_from_rows([
            bytearray(_row*times)
            for _row in self._rows
        ])

    def vtile(self, times=1):
        """Multiply height by row tiling."""
        return self._create_from_rows([
            bytearray(_row)
            for _ in range(times)
            for _row in self._rows
        ])

    def move(self, sy0, sy1, sx0, sx1, ty0, tx0):
        """Move a submatrix, replacing with attribute 0."""
        # copy or this won't work on a view
        clip = self[sy0:sy1, sx0:sx1].copy()
        height, width = sy1 - sy0, sx1 - sx0
        self[sy0:sy1, sx0:sx1] = 0
        self[ty0 : ty0+height, tx0 : tx0+width] = clip

    def to_bytes(self):
        """Convert to a bytes object (contiguous rows)."""
        # we need the bytearray cast - only in case we're a view
        return b''.join(bytes(bytearray(_row)) for _row in self._rows)

    def to_rows(self):
        """Convert to tuple of tuples of int."""
        return tuple(
            tuple(_i for _i in iterbytes(bytearray(_row)))
            for _row in self._rows
        )

    # views

    @property
    def view(self):
        """
        Create a bytematrixview of the current bytematrix.
        Use bm.view[yslice, xslice]
        """
        return self._create_from_rows([
            memoryview(_row) for _row in self._rows
        ])

    def copy(self):
        """
        Create a copy of the current bytematrix or view - as slicing views produces views.
        Use bm[yslice, xslice].copy()
        """
        return self._create_from_rows([
            bytearray(_row) for _row in self._rows
        ])

    @classmethod
    def view_from_buffer(cls, height, width, pitch, buffer):
        """Create a byte matrix as a view on a contiguous row-major buffer."""
        return cls._create_from_rows([
            memoryview(buffer)[_offset:_offset+width]
            for _offset in xrange(0, height*pitch, pitch)
        ])


##############################################################################
# concatenation

def hstack(matrices):
    """Horizontally concatenate matrices."""
    matrices = list(matrices)
    return ByteMatrix._create_from_rows([
        bytearray().join(_rows)
        for _rows in zip(*(_mat._rows for _mat in matrices))
    ])

def vstack(matrices):
    """Vertically concatenate matrices."""
    return ByteMatrix._create_from_rows([
        _row for _mat in matrices for _row in _mat._rows
    ])


##############################################################################
# bytearray functions

def unpack_bytes(packed, items_per_byte):
    """Unpack from packed-bits representation."""
    bpp = 8 // items_per_byte
    mask = (1 << bpp) - 1
    shifts = [8 - bpp - _sh for _sh in range(0, 8, bpp)]
    return bytearray(
        (_byte >> _shift) & mask
        for _byte in iterbytes(packed)
        for _shift in shifts
    )

def pack_bytes(unpacked, items_per_byte):
    """Pack into packed-bits representation."""
    bpp = 8 // items_per_byte
    mask = (1 << bpp) - 1
    shifts = [8 - bpp - _sh for _sh in range(0, 8, bpp)]
    # ceildiv(a,b) == -(floordiv(-a,b))
    packed_width = -(-len(unpacked) // items_per_byte)
    prepacked = [
        (_byte & mask) << _shift
        for _byte, _shift in zip(iterbytes(unpacked), shifts*packed_width)
    ]
    return bytearray([
        sum(prepacked[_offs : _offs+items_per_byte])
        for _offs in xrange(0, len(prepacked), items_per_byte)
    ])
